MEMO_MANAGER_ABI = [
  { "inputs": [], "stateMutability": "nonpayable", "type": "constructor" },
  { "inputs": [], "name": "AccessControlBadConfirmation", "type": "error" },
  {
    "inputs": [
      { "internalType": "address", "name": "account", "type": "address" },
      { "internalType": "bytes32", "name": "neededRole", "type": "bytes32" }
    ],
    "name": "AccessControlUnauthorizedAccount",
    "type": "error"
  },
  {
    "inputs": [{ "internalType": "address", "name": "target", "type": "address" }],
    "name": "AddressEmptyCode",
    "type": "error"
  },
  { "inputs": [], "name": "AlreadyVoted", "type": "error" },
  { "inputs": [], "name": "CannotApproveMemo", "type": "error" },
  { "inputs": [], "name": "CannotUpdateApprovedMemo", "type": "error" },
  { "inputs": [], "name": "CannotUpdateMemo", "type": "error" },
  { "inputs": [], "name": "CannotWithdrawYet", "type": "error" },
  { "inputs": [], "name": "DestinationChainNotConfigured", "type": "error" },
  {
    "inputs": [
      { "internalType": "address", "name": "implementation", "type": "address" }
    ],
    "name": "ERC1967InvalidImplementation",
    "type": "error"
  },
  { "inputs": [], "name": "ERC1967NonPayable", "type": "error" },
  { "inputs": [], "name": "EmptyContent", "type": "error" },
  { "inputs": [], "name": "FailedInnerCall", "type": "error" },
  { "inputs": [], "name": "InvalidInitialization", "type": "error" },
  { "inputs": [], "name": "InvalidMemoState", "type": "error" },
  { "inputs": [], "name": "InvalidMemoStateTransition", "type": "error" },
  { "inputs": [], "name": "InvalidMemoType", "type": "error" },
  { "inputs": [], "name": "JobAlreadyCompleted", "type": "error" },
  { "inputs": [], "name": "JobDoesNotExist", "type": "error" },
  { "inputs": [], "name": "MemoAlreadyApproved", "type": "error" },
  { "inputs": [], "name": "MemoAlreadyExecuted", "type": "error" },
  { "inputs": [], "name": "MemoAlreadySigned", "type": "error" },
  { "inputs": [], "name": "MemoCannotBeSigned", "type": "error" },
  { "inputs": [], "name": "MemoDoesNotExist", "type": "error" },
  { "inputs": [], "name": "MemoDoesNotRequireApproval", "type": "error" },
  { "inputs": [], "name": "MemoExpired", "type": "error" },
  { "inputs": [], "name": "MemoNotApproved", "type": "error" },
  { "inputs": [], "name": "MemoNotReadyToBeSigned", "type": "error" },
  { "inputs": [], "name": "MemoStateUnchanged", "type": "error" },
  { "inputs": [], "name": "NoAmountToTransfer", "type": "error" },
  { "inputs": [], "name": "NoPaymentAmount", "type": "error" },
  { "inputs": [], "name": "NotEscrowTransferMemoType", "type": "error" },
  { "inputs": [], "name": "NotInitializing", "type": "error" },
  { "inputs": [], "name": "NotPayableMemoType", "type": "error" },
  { "inputs": [], "name": "OnlyACPContract", "type": "error" },
  { "inputs": [], "name": "OnlyAssetManager", "type": "error" },
  { "inputs": [], "name": "OnlyClientOrProvider", "type": "error" },
  { "inputs": [], "name": "OnlyCounterParty", "type": "error" },
  { "inputs": [], "name": "OnlyEvaluator", "type": "error" },
  { "inputs": [], "name": "OnlyMemoSender", "type": "error" },
  { "inputs": [], "name": "ReentrancyGuardReentrantCall", "type": "error" },
  { "inputs": [], "name": "UUPSUnauthorizedCallContext", "type": "error" },
  {
    "inputs": [{ "internalType": "bytes32", "name": "slot", "type": "bytes32" }],
    "name": "UUPSUnsupportedProxiableUUID",
    "type": "error"
  },
  { "inputs": [], "name": "ZeroAcpContractAddress", "type": "error" },
  { "inputs": [], "name": "ZeroAddressRecipient", "type": "error" },
  { "inputs": [], "name": "ZeroAddressToken", "type": "error" },
  { "inputs": [], "name": "ZeroAssetManagerAddress", "type": "error" },
  { "inputs": [], "name": "ZeroJobManagerAddress", "type": "error" },
  {
    "anonymous": False,
    "inputs": [
      {
        "indexed": False,
        "internalType": "uint64",
        "name": "version",
        "type": "uint64"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "memoId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "address",
        "name": "approver",
        "type": "address"
      },
      { "indexed": False, "internalType": "bool", "name": "approved", "type": "bool" },
      {
        "indexed": False,
        "internalType": "string",
        "name": "reason",
        "type": "string"
      }
    ],
    "name": "MemoSigned",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "memoId",
        "type": "uint256"
      },
      {
        "indexed": False,
        "internalType": "enum ACPTypes.MemoState",
        "name": "oldState",
        "type": "uint8"
      },
      {
        "indexed": False,
        "internalType": "enum ACPTypes.MemoState",
        "name": "newState",
        "type": "uint8"
      }
    ],
    "name": "MemoStateUpdated",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "memoId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "jobId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "indexed": False,
        "internalType": "enum ACPTypes.MemoType",
        "name": "memoType",
        "type": "uint8"
      },
      {
        "indexed": False,
        "internalType": "enum ACPTypes.JobPhase",
        "name": "nextPhase",
        "type": "uint8"
      },
      {
        "indexed": False,
        "internalType": "string",
        "name": "content",
        "type": "string"
      }
    ],
    "name": "NewMemo",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "jobId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "memoId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "indexed": False,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": False,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "PayableFeeRefunded",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "jobId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "memoId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "indexed": False,
        "internalType": "address",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": False,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "PayableFundsRefunded",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "memoId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "uint256",
        "name": "jobId",
        "type": "uint256"
      },
      {
        "indexed": True,
        "internalType": "address",
        "name": "executor",
        "type": "address"
      },
      {
        "indexed": False,
        "internalType": "uint256",
        "name": "amount",
        "type": "uint256"
      }
    ],
    "name": "PayableMemoExecuted",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      { "indexed": True, "internalType": "bytes32", "name": "role", "type": "bytes32" },
      {
        "indexed": True,
        "internalType": "bytes32",
        "name": "previousAdminRole",
        "type": "bytes32"
      },
      {
        "indexed": True,
        "internalType": "bytes32",
        "name": "newAdminRole",
        "type": "bytes32"
      }
    ],
    "name": "RoleAdminChanged",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      { "indexed": True, "internalType": "bytes32", "name": "role", "type": "bytes32" },
      {
        "indexed": True,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": True,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "RoleGranted",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      { "indexed": True, "internalType": "bytes32", "name": "role", "type": "bytes32" },
      {
        "indexed": True,
        "internalType": "address",
        "name": "account",
        "type": "address"
      },
      {
        "indexed": True,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      }
    ],
    "name": "RoleRevoked",
    "type": "event"
  },
  {
    "anonymous": False,
    "inputs": [
      {
        "indexed": True,
        "internalType": "address",
        "name": "implementation",
        "type": "address"
      }
    ],
    "name": "Upgraded",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "ACP_CONTRACT_ROLE",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "ADMIN_ROLE",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "DEFAULT_ADMIN_ROLE",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "UPGRADE_INTERFACE_VERSION",
    "outputs": [{ "internalType": "string", "name": "", "type": "string" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "acpContract",
    "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "memoId", "type": "uint256" },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "bool", "name": "approved", "type": "bool" },
      { "internalType": "string", "name": "reason", "type": "string" }
    ],
    "name": "approveMemo",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "assetManager",
    "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256[]", "name": "memoIds", "type": "uint256[]" },
      { "internalType": "bool", "name": "approved", "type": "bool" },
      { "internalType": "string", "name": "reason", "type": "string" }
    ],
    "name": "bulkApproveMemos",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "memoId", "type": "uint256" },
      { "internalType": "address", "name": "user", "type": "address" }
    ],
    "name": "canApproveMemo",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "jobId", "type": "uint256" },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "string", "name": "content", "type": "string" },
      {
        "internalType": "enum ACPTypes.MemoType",
        "name": "memoType",
        "type": "uint8"
      },
      { "internalType": "bool", "name": "isSecured", "type": "bool" },
      {
        "internalType": "enum ACPTypes.JobPhase",
        "name": "nextPhase",
        "type": "uint8"
      },
      { "internalType": "string", "name": "metadata", "type": "string" }
    ],
    "name": "createMemo",
    "outputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "jobId", "type": "uint256" },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "string", "name": "content", "type": "string" },
      {
        "internalType": "enum ACPTypes.MemoType",
        "name": "memoType",
        "type": "uint8"
      },
      { "internalType": "bool", "name": "isSecured", "type": "bool" },
      {
        "internalType": "enum ACPTypes.JobPhase",
        "name": "nextPhase",
        "type": "uint8"
      },
      {
        "components": [
          { "internalType": "address", "name": "token", "type": "address" },
          { "internalType": "uint256", "name": "amount", "type": "uint256" },
          { "internalType": "address", "name": "recipient", "type": "address" },
          { "internalType": "uint256", "name": "feeAmount", "type": "uint256" },
          {
            "internalType": "enum ACPTypes.FeeType",
            "name": "feeType",
            "type": "uint8"
          },
          { "internalType": "bool", "name": "isExecuted", "type": "bool" },
          { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
          { "internalType": "uint32", "name": "lzSrcEid", "type": "uint32" },
          { "internalType": "uint32", "name": "lzDstEid", "type": "uint32" }
        ],
        "internalType": "struct ACPTypes.PayableDetails",
        "name": "payableDetails_",
        "type": "tuple"
      },
      { "internalType": "uint256", "name": "expiredAt", "type": "uint256" }
    ],
    "name": "createPayableMemo",
    "outputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "emergencyApproveMemo",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "executePayableMemo",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getAssetManager",
    "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "jobId", "type": "uint256" },
      { "internalType": "uint256", "name": "offset", "type": "uint256" },
      { "internalType": "uint256", "name": "limit", "type": "uint256" }
    ],
    "name": "getJobMemos",
    "outputs": [
      {
        "components": [
          { "internalType": "uint256", "name": "id", "type": "uint256" },
          { "internalType": "uint256", "name": "jobId", "type": "uint256" },
          { "internalType": "address", "name": "sender", "type": "address" },
          { "internalType": "string", "name": "content", "type": "string" },
          {
            "internalType": "enum ACPTypes.MemoType",
            "name": "memoType",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "createdAt", "type": "uint256" },
          { "internalType": "bool", "name": "isApproved", "type": "bool" },
          { "internalType": "address", "name": "approvedBy", "type": "address" },
          { "internalType": "uint256", "name": "approvedAt", "type": "uint256" },
          { "internalType": "bool", "name": "requiresApproval", "type": "bool" },
          { "internalType": "string", "name": "metadata", "type": "string" },
          { "internalType": "bool", "name": "isSecured", "type": "bool" },
          {
            "internalType": "enum ACPTypes.JobPhase",
            "name": "nextPhase",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
          {
            "internalType": "enum ACPTypes.MemoState",
            "name": "state",
            "type": "uint8"
          }
        ],
        "internalType": "struct ACPTypes.Memo[]",
        "name": "memoArray",
        "type": "tuple[]"
      },
      { "internalType": "uint256", "name": "total", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "jobId", "type": "uint256" },
      { "internalType": "enum ACPTypes.JobPhase", "name": "phase", "type": "uint8" },
      { "internalType": "uint256", "name": "offset", "type": "uint256" },
      { "internalType": "uint256", "name": "limit", "type": "uint256" }
    ],
    "name": "getJobMemosByPhase",
    "outputs": [
      {
        "components": [
          { "internalType": "uint256", "name": "id", "type": "uint256" },
          { "internalType": "uint256", "name": "jobId", "type": "uint256" },
          { "internalType": "address", "name": "sender", "type": "address" },
          { "internalType": "string", "name": "content", "type": "string" },
          {
            "internalType": "enum ACPTypes.MemoType",
            "name": "memoType",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "createdAt", "type": "uint256" },
          { "internalType": "bool", "name": "isApproved", "type": "bool" },
          { "internalType": "address", "name": "approvedBy", "type": "address" },
          { "internalType": "uint256", "name": "approvedAt", "type": "uint256" },
          { "internalType": "bool", "name": "requiresApproval", "type": "bool" },
          { "internalType": "string", "name": "metadata", "type": "string" },
          { "internalType": "bool", "name": "isSecured", "type": "bool" },
          {
            "internalType": "enum ACPTypes.JobPhase",
            "name": "nextPhase",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
          {
            "internalType": "enum ACPTypes.MemoState",
            "name": "state",
            "type": "uint8"
          }
        ],
        "internalType": "struct ACPTypes.Memo[]",
        "name": "memoArray",
        "type": "tuple[]"
      },
      { "internalType": "uint256", "name": "total", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "jobId", "type": "uint256" },
      {
        "internalType": "enum ACPTypes.MemoType",
        "name": "memoType",
        "type": "uint8"
      },
      { "internalType": "uint256", "name": "offset", "type": "uint256" },
      { "internalType": "uint256", "name": "limit", "type": "uint256" }
    ],
    "name": "getJobMemosByType",
    "outputs": [
      {
        "components": [
          { "internalType": "uint256", "name": "id", "type": "uint256" },
          { "internalType": "uint256", "name": "jobId", "type": "uint256" },
          { "internalType": "address", "name": "sender", "type": "address" },
          { "internalType": "string", "name": "content", "type": "string" },
          {
            "internalType": "enum ACPTypes.MemoType",
            "name": "memoType",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "createdAt", "type": "uint256" },
          { "internalType": "bool", "name": "isApproved", "type": "bool" },
          { "internalType": "address", "name": "approvedBy", "type": "address" },
          { "internalType": "uint256", "name": "approvedAt", "type": "uint256" },
          { "internalType": "bool", "name": "requiresApproval", "type": "bool" },
          { "internalType": "string", "name": "metadata", "type": "string" },
          { "internalType": "bool", "name": "isSecured", "type": "bool" },
          {
            "internalType": "enum ACPTypes.JobPhase",
            "name": "nextPhase",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
          {
            "internalType": "enum ACPTypes.MemoState",
            "name": "state",
            "type": "uint8"
          }
        ],
        "internalType": "struct ACPTypes.Memo[]",
        "name": "memoArray",
        "type": "tuple[]"
      },
      { "internalType": "uint256", "name": "total", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getLocalEid",
    "outputs": [{ "internalType": "uint32", "name": "", "type": "uint32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "getMemo",
    "outputs": [
      {
        "components": [
          { "internalType": "uint256", "name": "id", "type": "uint256" },
          { "internalType": "uint256", "name": "jobId", "type": "uint256" },
          { "internalType": "address", "name": "sender", "type": "address" },
          { "internalType": "string", "name": "content", "type": "string" },
          {
            "internalType": "enum ACPTypes.MemoType",
            "name": "memoType",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "createdAt", "type": "uint256" },
          { "internalType": "bool", "name": "isApproved", "type": "bool" },
          { "internalType": "address", "name": "approvedBy", "type": "address" },
          { "internalType": "uint256", "name": "approvedAt", "type": "uint256" },
          { "internalType": "bool", "name": "requiresApproval", "type": "bool" },
          { "internalType": "string", "name": "metadata", "type": "string" },
          { "internalType": "bool", "name": "isSecured", "type": "bool" },
          {
            "internalType": "enum ACPTypes.JobPhase",
            "name": "nextPhase",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
          {
            "internalType": "enum ACPTypes.MemoState",
            "name": "state",
            "type": "uint8"
          }
        ],
        "internalType": "struct ACPTypes.Memo",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "getMemoApprovalStatus",
    "outputs": [
      { "internalType": "bool", "name": "isApproved", "type": "bool" },
      { "internalType": "address", "name": "approvedBy", "type": "address" },
      { "internalType": "uint256", "name": "approvedAt", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "getMemoWithPayableDetails",
    "outputs": [
      {
        "components": [
          { "internalType": "uint256", "name": "id", "type": "uint256" },
          { "internalType": "uint256", "name": "jobId", "type": "uint256" },
          { "internalType": "address", "name": "sender", "type": "address" },
          { "internalType": "string", "name": "content", "type": "string" },
          {
            "internalType": "enum ACPTypes.MemoType",
            "name": "memoType",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "createdAt", "type": "uint256" },
          { "internalType": "bool", "name": "isApproved", "type": "bool" },
          { "internalType": "address", "name": "approvedBy", "type": "address" },
          { "internalType": "uint256", "name": "approvedAt", "type": "uint256" },
          { "internalType": "bool", "name": "requiresApproval", "type": "bool" },
          { "internalType": "string", "name": "metadata", "type": "string" },
          { "internalType": "bool", "name": "isSecured", "type": "bool" },
          {
            "internalType": "enum ACPTypes.JobPhase",
            "name": "nextPhase",
            "type": "uint8"
          },
          { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
          {
            "internalType": "enum ACPTypes.MemoState",
            "name": "state",
            "type": "uint8"
          }
        ],
        "internalType": "struct ACPTypes.Memo",
        "name": "memo",
        "type": "tuple"
      },
      {
        "components": [
          { "internalType": "address", "name": "token", "type": "address" },
          { "internalType": "uint256", "name": "amount", "type": "uint256" },
          { "internalType": "address", "name": "recipient", "type": "address" },
          { "internalType": "uint256", "name": "feeAmount", "type": "uint256" },
          {
            "internalType": "enum ACPTypes.FeeType",
            "name": "feeType",
            "type": "uint8"
          },
          { "internalType": "bool", "name": "isExecuted", "type": "bool" },
          { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
          { "internalType": "uint32", "name": "lzSrcEid", "type": "uint32" },
          { "internalType": "uint32", "name": "lzDstEid", "type": "uint32" }
        ],
        "internalType": "struct ACPTypes.PayableDetails",
        "name": "details",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "getPayableDetails",
    "outputs": [
      {
        "components": [
          { "internalType": "address", "name": "token", "type": "address" },
          { "internalType": "uint256", "name": "amount", "type": "uint256" },
          { "internalType": "address", "name": "recipient", "type": "address" },
          { "internalType": "uint256", "name": "feeAmount", "type": "uint256" },
          {
            "internalType": "enum ACPTypes.FeeType",
            "name": "feeType",
            "type": "uint8"
          },
          { "internalType": "bool", "name": "isExecuted", "type": "bool" },
          { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
          { "internalType": "uint32", "name": "lzSrcEid", "type": "uint32" },
          { "internalType": "uint32", "name": "lzDstEid", "type": "uint32" }
        ],
        "internalType": "struct ACPTypes.PayableDetails",
        "name": "",
        "type": "tuple"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "bytes32", "name": "role", "type": "bytes32" }],
    "name": "getRoleAdmin",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "grantRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "hasRole",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "acpContract_", "type": "address" },
      { "internalType": "address", "name": "jobManager_", "type": "address" },
      { "internalType": "address", "name": "paymentManager_", "type": "address" }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "jobId", "type": "uint256" },
      { "internalType": "address", "name": "user", "type": "address" }
    ],
    "name": "isJobEvaluator",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "memoId", "type": "uint256" },
      { "internalType": "address", "name": "user", "type": "address" }
    ],
    "name": "isMemoSigner",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "isPayable",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "jobManager",
    "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "", "type": "uint256" },
      { "internalType": "uint256", "name": "", "type": "uint256" }
    ],
    "name": "jobMemos",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "", "type": "uint256" },
      { "internalType": "enum ACPTypes.JobPhase", "name": "", "type": "uint8" },
      { "internalType": "uint256", "name": "", "type": "uint256" }
    ],
    "name": "jobMemosByPhase",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "", "type": "uint256" },
      { "internalType": "enum ACPTypes.MemoType", "name": "", "type": "uint8" },
      { "internalType": "uint256", "name": "", "type": "uint256" }
    ],
    "name": "jobMemosByType",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "", "type": "uint256" },
      { "internalType": "address", "name": "", "type": "address" }
    ],
    "name": "memoApprovals",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "memoCounter",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "name": "memos",
    "outputs": [
      { "internalType": "uint256", "name": "id", "type": "uint256" },
      { "internalType": "uint256", "name": "jobId", "type": "uint256" },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "string", "name": "content", "type": "string" },
      {
        "internalType": "enum ACPTypes.MemoType",
        "name": "memoType",
        "type": "uint8"
      },
      { "internalType": "uint256", "name": "createdAt", "type": "uint256" },
      { "internalType": "bool", "name": "isApproved", "type": "bool" },
      { "internalType": "address", "name": "approvedBy", "type": "address" },
      { "internalType": "uint256", "name": "approvedAt", "type": "uint256" },
      { "internalType": "bool", "name": "requiresApproval", "type": "bool" },
      { "internalType": "string", "name": "metadata", "type": "string" },
      { "internalType": "bool", "name": "isSecured", "type": "bool" },
      {
        "internalType": "enum ACPTypes.JobPhase",
        "name": "nextPhase",
        "type": "uint8"
      },
      { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
      { "internalType": "enum ACPTypes.MemoState", "name": "state", "type": "uint8" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "name": "payableDetails",
    "outputs": [
      { "internalType": "address", "name": "token", "type": "address" },
      { "internalType": "uint256", "name": "amount", "type": "uint256" },
      { "internalType": "address", "name": "recipient", "type": "address" },
      { "internalType": "uint256", "name": "feeAmount", "type": "uint256" },
      { "internalType": "enum ACPTypes.FeeType", "name": "feeType", "type": "uint8" },
      { "internalType": "bool", "name": "isExecuted", "type": "bool" },
      { "internalType": "uint256", "name": "expiredAt", "type": "uint256" },
      { "internalType": "uint32", "name": "lzSrcEid", "type": "uint32" },
      { "internalType": "uint32", "name": "lzDstEid", "type": "uint32" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "paymentManager",
    "outputs": [{ "internalType": "address", "name": "", "type": "address" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "proxiableUUID",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" },
      { "internalType": "address", "name": "callerConfirmation", "type": "address" }
    ],
    "name": "renounceRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "name": "requiredApprovals",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "requiresApproval",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "role", "type": "bytes32" },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "revokeRole",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "enum ACPTypes.MemoType",
        "name": "memoType",
        "type": "uint8"
      },
      { "internalType": "uint256", "name": "requiredApprovals_", "type": "uint256" }
    ],
    "name": "setApprovalRequirements",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "assetManager_", "type": "address" }
    ],
    "name": "setAssetManager",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "memoId", "type": "uint256" },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "bool", "name": "isApproved", "type": "bool" },
      { "internalType": "string", "name": "reason", "type": "string" }
    ],
    "name": "signMemo",
    "outputs": [{ "internalType": "uint256", "name": "jobId", "type": "uint256" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "bytes4", "name": "interfaceId", "type": "bytes4" }],
    "name": "supportsInterface",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "acpContract_", "type": "address" },
      { "internalType": "address", "name": "jobManager_", "type": "address" },
      { "internalType": "address", "name": "paymentManager_", "type": "address" },
      { "internalType": "address", "name": "assetManager_", "type": "address" }
    ],
    "name": "updateContracts",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "memoId", "type": "uint256" },
      { "internalType": "string", "name": "newContent", "type": "string" }
    ],
    "name": "updateMemoContent",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "memoId", "type": "uint256" },
      {
        "internalType": "enum ACPTypes.MemoState",
        "name": "newMemoState",
        "type": "uint8"
      }
    ],
    "name": "updateMemoState",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "newImplementation", "type": "address" },
      { "internalType": "bytes", "name": "data", "type": "bytes" }
    ],
    "name": "upgradeToAndCall",
    "outputs": [],
    "stateMutability": "payable",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "uint256", "name": "memoId", "type": "uint256" }],
    "name": "withdrawEscrowedFunds",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  }
];


import time
import requests
import secrets
from typing import Any, Dict, Optional
from eth_account.messages import encode_defunct


from virtuals_acp.constants import (
    HTTP_STATUS_CODES_X402,
    SINGLE_SIGNER_VALIDATION_MODULE_ADDRESS,
    X402_AUTHORIZATION_TYPES,
)
from virtuals_acp.models import (
    X402PayableRequest,
    X402PayableRequirements,
    X402Payment,
    OffChainJob,
    X402PaymentPayload,
)
from virtuals_acp.exceptions import ACPError
from virtuals_acp.configs.configs import ACPContractConfig
from virtuals_acp.abis.erc20_abi import ERC20_ABI
from virtuals_acp.abis.flat_token_v2_abi import FIAT_TOKEN_V2_ABI
from eth_account.messages import encode_typed_data
from eth_utils.crypto import keccak

from virtuals_acp.utils import safe_base64_encode


class ACPX402:
    def __init__(
        self,
        config: ACPContractConfig,
        session_key_client,
        public_client,
        agent_wallet_address: str,
        entity_id: int,
    ):
        """
        config: ACPContractConfig
        session_key_client: a client capable of signing messages and typed data
        public_client: web3 client able to read from contracts
        """
        self.config = config
        self.session_key_client = session_key_client
        self.public_client = public_client
        self.agent_wallet_address = agent_wallet_address
        self.entity_id = entity_id

    def sign_update_job_nonce_message(self, job_id: int, nonce: str) -> str:
        message = f"{job_id}-{nonce}"
        try:
            signature = self.session_key_client.sign_message(
                encode_defunct(text=message)
            )
            return signature
        except Exception as e:
            raise ACPError("Failed to sign update job X402 nonce message", e)

    def update_job_nonce(self, job_id: int, nonce: str) -> OffChainJob:
        """Update job X402 nonce."""
        try:
            api_url = f"{self.config.acp_api_url}/jobs/{job_id}/x402-nonce"
            message = f"{job_id}-{nonce}"

            # Use eth_account.messages.encode_defunct to encode the message as an EIP-191 message (Ethereum signed message)
            eth_message = encode_defunct(text=message)
            signature = self.session_key_client.sign_message(eth_message)

            headers = {
                "x-signature": "0x" + signature.signature.hex(),
                "x-nonce": nonce,
                "Content-Type": "application/json",
            }

            payload = {"data": {"nonce": nonce}}

            response = requests.post(api_url, headers=headers, json=payload)

            if not response.ok:
                raise ACPError("Failed to update job X402 nonce", response.text)
            return response.json()
        except Exception as e:
            raise ACPError("Failed to update job X402 nonce", e)

    def generate_payment(
        self, payable_request: X402PayableRequest, requirements: X402PayableRequirements
    ) -> X402Payment:
        try:
            usdc_contract = self.config.base_fare.contract_address
            time_now = int(time.time())
            valid_after = str(time_now - 60)
            valid_before = str(time_now + requirements.accepts[0].maxTimeoutSeconds)

            # Get token name and version using multicall but python not supported
            usdc_contract_instance = self.public_client.eth.contract(
                address=usdc_contract, abi=ERC20_ABI
            )

            token_name = usdc_contract_instance.functions.name().call()

            # Get version from FIAT_TOKEN_V2_ABI
            fiat_token_contract = self.public_client.eth.contract(
                address=usdc_contract, abi=FIAT_TOKEN_V2_ABI
            )
            token_version = fiat_token_contract.functions.version().call()

            nonce_bytes = secrets.token_bytes(32)
            nonce = "0x" + nonce_bytes.hex()

            message = {
                "from": self.agent_wallet_address,
                "to": payable_request.to,
                "value": str(payable_request.value),
                "validAfter": valid_after,
                "validBefore": valid_before,
                "nonce": nonce,
            }

            types = {
                "TransferWithAuthorization": X402_AUTHORIZATION_TYPES,
            }

            domain = {
                "name": str(token_name),
                "version": str(token_version),
                "chainId": int(self.config.chain_id),
                "verifyingContract": str(usdc_contract),
            }

            encoded_typed_data = encode_typed_data(
                full_message={
                    "types": types,
                    "domain": domain,
                    "message": message,
                    "primaryType": "TransferWithAuthorization",
                }
            )

            typed_data_hash = keccak(
                b"\x19\x01" + encoded_typed_data.header + encoded_typed_data.body
            )

            replay_safe_typed_data = {
                "domain": {
                    "chainId": int(self.config.chain_id),
                    "verifyingContract": SINGLE_SIGNER_VALIDATION_MODULE_ADDRESS,
                    "salt": "0x"
                    + "00" * 12
                    + self.agent_wallet_address[
                        2:
                    ],  # Assuming account_address is '0x...'
                },
                "types": {"ReplaySafeHash": [{"name": "hash", "type": "bytes32"}]},
                "message": {"hash": "0x" + typed_data_hash.hex()},
                "primaryType": "ReplaySafeHash",
            }

            signed_msg = self.session_key_client.sign_typed_data(
                full_message=replay_safe_typed_data
            )

            raw_signature = signed_msg.signature.hex()

            if not raw_signature.startswith("0x"):
                raw_signature = "0x" + raw_signature

            final_signature = self.pack_1271_eoa_signature(
                raw_signature, self.entity_id
            )

            payload = X402PaymentPayload(
                x402_version=requirements.x402Version,
                scheme=requirements.accepts[0].scheme,
                network=requirements.accepts[0].network,
                payload={
                    "signature": final_signature,
                    "authorization": message
                }
            )

            encoded_payment = self.encode_payment(payload)

            return X402Payment(
                encodedPayment=encoded_payment,
                message=message,
                signature=final_signature,
            )

        except Exception as error:
            raise ACPError("Failed to generate X402 payment", error)

    def perform_request(
        self, url: str, version: str, budget: Optional[str] = None, signature: Optional[str] = None
    ) -> Dict[str, Any]:
        base_url = self.config.x402_config.url if self.config.x402_config else None

        if not base_url:
            raise ACPError("X402 URL not configured")

        try:
            headers = {}
            if signature:
                headers["x-payment"] = signature
            if budget:
                headers["x-budget"] = str(budget)
                
            headers["x-acp-version"] = version

            res = requests.get(f"{base_url}{url}", headers=headers, timeout=60)
            data = res.json()                    
            
            if not res.ok and res.status_code != HTTP_STATUS_CODES_X402["Payment Required"]:
                raise ACPError("Invalid response status code for X402 request", data)
            
            return {
                "isPaymentRequired": res.status_code == HTTP_STATUS_CODES_X402["Payment Required"],
                "data": data
            }
        except Exception as error:
            raise ACPError("Failed to perform X402 request", error)

    def encode_payment(self, payment_payload: X402PaymentPayload) -> str:
        return safe_base64_encode(payment_payload.model_dump_json())

    def pack_1271_eoa_signature(self, validation_signature: str, entity_id: int) -> str:
        if not validation_signature.startswith("0x"):
            validation_signature = "0x" + validation_signature

        # Components
        prefix = b"\x00"  # 0x00
        entity_id_bytes = entity_id.to_bytes(4, "big")  # 4 bytes
        separator = b"\xff"  # 0xFF
        eoa_type = b"\x00"  # 0x00 (EOA type)
        sig_bytes = bytes.fromhex(validation_signature[2:])

        # Concatenate all parts
        packed = prefix + entity_id_bytes + separator + eoa_type + sig_bytes

        return "0x" + packed.hex()



import json
import re
import time
from datetime import datetime, timezone, timedelta
from typing import TYPE_CHECKING, List, Optional, Dict, Any, Union, Literal
from pydantic import BaseModel, Field, ConfigDict, PrivateAttr

from virtuals_acp.account import ACPAccount
from virtuals_acp.exceptions import ACPError
from virtuals_acp.memo import ACPMemo
from virtuals_acp.models import (
    ACPMemoState,
    PriceType,
    OperationPayload,
    X402PayableRequest,
    X402PayableRequirements,
    RequestPayload,
)
from virtuals_acp.utils import (
    get_destination_chain_id,
    get_destination_endpoint_id,
    try_parse_json_model,
    prepare_payload,
    get_txn_hash_from_response,
)
from virtuals_acp.models import (
    ACPJobPhase,
    MemoType,
    IACPAgent,
    DeliverablePayload,
    FeeType,
)
from virtuals_acp.fare import Fare, FareAmountBase, FareAmount
from virtuals_acp.web3 import getERC20Allowance, getERC20Balance, getERC20Decimals, getERC20Symbol

if TYPE_CHECKING:
    from virtuals_acp.client import VirtualsACP


class ACPJob(BaseModel):
    acp_client: "VirtualsACP"
    id: int
    client_address: str
    provider_address: str
    evaluator_address: str
    price: float
    price_token_address: Optional[str] = None
    memos: List[ACPMemo] = Field(default_factory=list)
    phase: ACPJobPhase
    context: Dict[str, Any] | None
    contract_address: Optional[str] = None
    net_payable_amount: Optional[float] = None
    deliverable: Optional[DeliverablePayload] = None # TODO: turn this into private attr

    model_config = ConfigDict(arbitrary_types_allowed=True)

    _base_fare: Optional[Fare] = PrivateAttr(default=None)
    _name: Optional[str] = PrivateAttr(default=None)
    _requirement: Optional[Union[str, Dict[str, Any]]] = PrivateAttr(default=None)
    _price_type: PriceType = PrivateAttr(default=PriceType.FIXED)
    _price_value: float = PrivateAttr(default=0.0)

    def model_post_init(self, __context: Any) -> None:
        if self.acp_client:
            self._base_fare = self.acp_client.acp_contract_client.config.base_fare

        memo = next(
            (
                m
                for m in self.memos
                if ACPJobPhase(m.next_phase) == ACPJobPhase.NEGOTIATION
            ),
            None,
        )

        if not memo:
            return None

        if not memo.content:
            return None

        content_obj = try_parse_json_model(memo.content, RequestPayload)

        if not content_obj:
            return None

        self._requirement = content_obj.service_requirement or content_obj.requirement
        self._name = content_obj.service_name or content_obj.name
        self._price_type = content_obj.price_type or PriceType.FIXED
        self._price_value = content_obj.price_value or self.price
        return None

    @property
    def requirement(self) -> Optional[Union[str, Dict[str, Any]]]:
        return self._requirement

    @property
    def name(self) -> Optional[str]:
        return self._name

    @property
    def price_type(self) -> PriceType:
        return self._price_type

    @property
    def price_value(self) -> float:
        return self._price_value

    def __str__(self):
        return (
            f"AcpJob(\n"
            f"  id={self.id},\n"
            f"  client_address='{self.client_address}',\n"
            f"  provider_address='{self.provider_address}',\n"
            f"  evaluator_address='{self.evaluator_address}',\n"
            f"  price={self.price},\n"
            f"  price_token_address='{self.price_token_address}',\n"
            f"  memos=[{', '.join(str(memo) for memo in self.memos)}],\n"
            f"  phase={self.phase}\n"
            f"  context={self.context}\n"
            f"  contract_address='{self.contract_address}',\n"
            f"  net_payable_amount='{self.net_payable_amount}',\n"
            f")"
        )

    @property
    def acp_contract_client(self):
        if not self.contract_address:
            return self.acp_client.acp_contract_client
        return self.acp_client.contract_client_by_address(self.contract_address)

    @property
    def config(self):
        return self.acp_contract_client.config

    @property
    def base_fare(self) -> Fare:
        return self.acp_contract_client.config.base_fare

    @property
    def account(self) -> Optional[ACPAccount]:
        return self.acp_client.get_account_by_job_id(self.id, self.acp_contract_client)


    @property
    def rejection_reason(self) -> Optional[str]:
        """Get the rejection reason from the rejected memo"""
        if self.phase != ACPJobPhase.REJECTED:
            return None

        request_memo = next(
            (m for m in self.memos if m.next_phase == ACPJobPhase.NEGOTIATION),
            None,
        )
        if request_memo:
            return request_memo.signed_reason

        fallback_memo = next(
            (m for m in self.memos if m.next_phase == ACPJobPhase.REJECTED), None
        )
        return fallback_memo.content if fallback_memo else None

    def create_requirement(self, content: str) -> str | None:
        operations: List[OperationPayload] = []

        operations.append(
            self.acp_contract_client.create_memo(
                job_id=self.id,
                content=content,
                memo_type=MemoType.MESSAGE,
                is_secured=False,
                next_phase=ACPJobPhase.TRANSACTION,
            )
        )

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def create_payable_requirement(
        self,
        content: str,
        type: Literal[MemoType.PAYABLE_REQUEST, MemoType.PAYABLE_TRANSFER_ESCROW],
        amount: FareAmountBase,
        recipient: str,
        expired_at: Optional[datetime] = None,
    ) -> str | None:
        operations: List[OperationPayload] = []

        if expired_at is None:
            expired_at = datetime.now(timezone.utc) + timedelta(minutes=5)
        if type == MemoType.PAYABLE_TRANSFER_ESCROW:
            operations.append(
                self.acp_contract_client.approve_allowance(
                    amount.amount,
                    amount.fare.contract_address,
                )
            )

        if self._price_type == PriceType.PERCENTAGE:
            fee_amount = int(self.price_value * 10000)
            fee_type = FeeType.PERCENTAGE_FEE
        else:
            fee_amount = (FareAmount(0, self.base_fare)).amount
            fee_type = FeeType.NO_FEE

        if amount.fare.chain_id != self.acp_contract_client.config.chain_id:
            operations.append(
                self.acp_contract_client.create_cross_chain_payable_memo(
                    job_id=self.id,
                    content=content,
                    amount_base_unit=amount.amount,
                    recipient=recipient,
                    fee_amount_base_unit=fee_amount,
                    fee_type=fee_type,
                    token=amount.fare.contract_address,
                    next_phase=ACPJobPhase.TRANSACTION,
                    memo_type=type,
                    expired_at=expired_at,
                    destination_eid=get_destination_endpoint_id(amount.fare.chain_id),
                )
            )
        else:
            operations.append(
                self.acp_contract_client.create_payable_memo(
                    job_id=self.id,
                    content=content,
                    amount_base_unit=amount.amount,
                    recipient=recipient,
                    fee_amount_base_unit=fee_amount,
                    fee_type=fee_type,
                    next_phase=ACPJobPhase.TRANSACTION,
                    memo_type=type,
                    expired_at=expired_at,
                    token=amount.fare.contract_address,
                )
            )

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def pay_and_accept_requirement(self, reason: Optional[str] = "") -> str | None:
        memo = next(
            (m for m in self.memos if m.next_phase == ACPJobPhase.TRANSACTION), None
        )

        if not memo:
            raise Exception("No negotiation memo found")

        if (
            memo.type == MemoType.PAYABLE_REQUEST and
            memo.state != ACPMemoState.PENDING and
            memo.payable_details is not None and
            memo.payable_details.get('lzDstEid') is not None
        ):
            print(f"Memo not ready to be signed, state: {memo.state}, payable_details: {memo.payable_details}")
            return

        if memo.payable_details:
            if "lzDstEid" in memo.payable_details:
                destination_chain_id = get_destination_chain_id(memo.payable_details["lzDstEid"])
            else:
                destination_chain_id = self.acp_contract_client.config.chain_id

            if(destination_chain_id != self.acp_contract_client.config.chain_id):
                token_balance = getERC20Balance(
                    self.acp_contract_client.public_clients[destination_chain_id],
                    memo.payable_details["token"],
                    self.client_address,
                )

                if token_balance < memo.payable_details["amount"]:
                    token_decimals = getERC20Decimals(
                        self.acp_contract_client.public_clients[destination_chain_id],
                        memo.payable_details["token"],
                    )

                    token_symbol = getERC20Symbol(
                        self.acp_contract_client.public_clients[destination_chain_id],
                        memo.payable_details["token"],
                    )

                    raise ACPError(f"You do not have enough funds to pay for the job which costs {memo.payable_details['amount'] / 10 ** token_decimals} {token_symbol} on chainId {destination_chain_id}")
                else:
                    asset_manager_address = self.acp_contract_client.get_asset_manager_address()

                    allowance = getERC20Allowance(
                        self.acp_contract_client.public_clients[destination_chain_id],
                        memo.payable_details["token"],
                        self.client_address,
                        asset_manager_address,
                    )

                    destination_chain_operations: List[OperationPayload] = []

                    destination_chain_operations.append(
                        self.acp_contract_client.approve_allowance(
                            memo.payable_details["amount"] + allowance,
                            memo.payable_details["token"],
                            asset_manager_address,
                        )
                    )

                    self.acp_contract_client.handle_operation(destination_chain_operations, destination_chain_id)

        operations: List[OperationPayload] = []
        base_fare_amount = FareAmount(self.price, self.base_fare)

        if memo.payable_details:
            transfer_amount = FareAmountBase.from_contract_address(
                memo.payable_details.get("amount"),
                memo.payable_details.get("token"),
                self.config,
            )
        else:
            transfer_amount = FareAmount(0, self.base_fare)

        # merge amounts if same token
        if (
            base_fare_amount.fare.contract_address
            == transfer_amount.fare.contract_address
        ):
            total_amount = base_fare_amount.add(transfer_amount)
        else:
            total_amount = base_fare_amount

        # approve base fare
        operations.append(
            self.acp_contract_client.approve_allowance(
                total_amount.amount,
                self.base_fare.contract_address,
            )
        )

        # approve transfer if token differs
        if (
            base_fare_amount.fare.contract_address
            != transfer_amount.fare.contract_address
        ):
            operations.append(
                self.acp_contract_client.approve_allowance(
                    transfer_amount.amount,
                    transfer_amount.fare.contract_address,
                )
            )

        # sign memo
        operations.append(self.acp_contract_client.sign_memo(memo.id, True, reason))

        operations.append(
            self.acp_contract_client.create_memo(
                job_id=self.id,
                content=f"Payment made. {reason or ''}".strip(),
                memo_type=MemoType.MESSAGE,
                is_secured=True,
                next_phase=ACPJobPhase.EVALUATION,
            )
        )

        if self.price > 0:
            x402PaymentDetails = self.acp_contract_client.get_x402_payment_details(
                self.id
            )
            if x402PaymentDetails.is_x402:
                self.perform_x402_payment(self.price)

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def accept(self, reason: Optional[str] = None) -> str | None:
        memo_content = f"Job {self.id} accepted. {reason or ''}"
        latest_memo = self.latest_memo
        if latest_memo is None or latest_memo.next_phase != ACPJobPhase.NEGOTIATION:
            raise ValueError("No request memo found")

        return latest_memo.sign(True, memo_content)

    def reject(self, reason: Optional[str] = None) -> str | None:
        memo_content = f"Job {self.id} rejected. {reason or ''}"
        latest_memo = self.latest_memo
        operations: List[OperationPayload] = []

        if self.phase is ACPJobPhase.REQUEST:
            if latest_memo is None or latest_memo.next_phase != ACPJobPhase.NEGOTIATION:
                raise ValueError("No request memo found")

            return latest_memo.sign(False, memo_content)

        operations.append(
            self.acp_contract_client.create_memo(
                job_id=self.id,
                content=memo_content,
                memo_type=MemoType.MESSAGE,
                is_secured=True,
                next_phase=ACPJobPhase.REJECTED,
            )
        )

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def reject_payable(
        self,
        reason: Optional[str],
        amount: FareAmountBase,
        expired_at: Optional[datetime] = None,
    ) -> str | None:
        if expired_at is None:
            expired_at = datetime.now(timezone.utc) + timedelta(minutes=5)

        memo_content = f"Job {self.id} rejected. {reason or ''}"
        fee_amount = FareAmount(0, self.acp_contract_client.config.base_fare)
        operations: List[OperationPayload] = []

        operations.append(
            self.acp_contract_client.approve_allowance(
                amount.amount, amount.fare.contract_address
            )
        )

        operations.append(
            self.acp_contract_client.create_payable_memo(
                job_id=self.id,
                content=memo_content,
                amount_base_unit=amount.amount,
                recipient=self.client_address,
                fee_amount_base_unit=fee_amount.amount,
                fee_type=FeeType.NO_FEE,
                next_phase=ACPJobPhase.REJECTED,
                memo_type=MemoType.PAYABLE_TRANSFER,
                expired_at=expired_at,
                token=amount.fare.contract_address,
            )
        )

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def respond(
        self,
        accept: bool,
        reason: Optional[str] = None,
    ) -> str | None:
        memo_content = (
            f"Job {self.id} {'accepted' if accept else 'rejected'}. {reason or ''}"
        )
        if accept:
            self.accept(memo_content)
            return self.create_requirement(memo_content)

        return self.reject(memo_content)

    @property
    def provider_agent(self) -> Optional["IACPAgent"]:
        """Get the provider agent details"""
        return self.acp_client.get_agent(self.provider_address)

    @property
    def client_agent(self) -> Optional["IACPAgent"]:
        """Get the client agent details"""
        return self.acp_client.get_agent(self.client_address)

    @property
    def evaluator_agent(self) -> Optional["IACPAgent"]:
        """Get the evaluator agent details"""
        return self.acp_client.get_agent(self.evaluator_address)

    @property
    def latest_memo(self) -> Optional[ACPMemo]:
        """Get the latest memo in the job"""
        return self.memos[-1] if self.memos else None

    def _get_memo_by_id(self, memo_id: int) -> Optional[ACPMemo]:
        return next((m for m in self.memos if m.id == memo_id), None)

    def deliver(self, deliverable: DeliverablePayload) -> str | None:
        if self.phase != ACPJobPhase.TRANSACTION:
            raise ACPError("Job is not in transaction phase")
        
        operations: List[OperationPayload] = []

        operations.append(
            self.acp_contract_client.create_memo(
                job_id=self.id,
                content=prepare_payload(deliverable),
                memo_type=MemoType.MESSAGE,
                is_secured=True,
                next_phase=ACPJobPhase.COMPLETED,
            )
        )

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def deliver_payable(
        self,
        deliverable: DeliverablePayload,
        amount: FareAmountBase,
        skip_fee: bool = False,
        expired_at: Optional[datetime] = None,
    ) -> str | None:
        if self.phase != ACPJobPhase.TRANSACTION:
            raise ACPError("Job is not in transaction phase")

        if expired_at is None:
            expired_at = datetime.now(timezone.utc) + timedelta(minutes=5)

        if amount.fare.chain_id != self.acp_contract_client.config.chain_id:
            return self._deliver_cross_chain_payable(self.client_address, amount)

        operations: List[OperationPayload] = []

        operations.append(
            self.acp_contract_client.approve_allowance(
                amount.amount, amount.fare.contract_address
            )
        )

        if self._price_type == PriceType.PERCENTAGE and not skip_fee:
            fee_amount = int(self.price_value * 10000)
            fee_type = FeeType.PERCENTAGE_FEE
        else:
            fee_amount = (FareAmount(0, self.base_fare)).amount
            fee_type = FeeType.NO_FEE

        operations.append(
            self.acp_contract_client.create_payable_memo(
                job_id=self.id,
                content=prepare_payload(deliverable),
                amount_base_unit=amount.amount,
                recipient=self.client_address,
                fee_amount_base_unit=fee_amount,
                fee_type=fee_type,
                next_phase=ACPJobPhase.COMPLETED,
                memo_type=MemoType.PAYABLE_TRANSFER,
                expired_at=expired_at,
                token=amount.fare.contract_address,
            )
        )

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def evaluate(self, accept: bool, reason: Optional[str] = None) -> str | None:
        if (
            self.latest_memo is None
            or self.latest_memo.next_phase != ACPJobPhase.COMPLETED
        ):
            raise ValueError("No evaluation memo found")

        if not reason:
            reason = f"Job {self.id} delivery {'accepted' if accept else 'rejected'}"

        return self.latest_memo.sign(accept, reason)

    def create_notification(self, content: str) -> str | None:
        operations: List[OperationPayload] = []

        operations.append(
            self.acp_contract_client.create_memo(
                job_id=self.id,
                content=content,
                memo_type=MemoType.NOTIFICATION,
                is_secured=True,
                next_phase=ACPJobPhase.COMPLETED,
            )
        )

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def create_payable_notification(
        self,
        content: str,
        amount: FareAmountBase,
        skip_fee: bool = False,
        expired_at: Optional[datetime] = None,
    ):
        operations: List[OperationPayload] = []

        if expired_at is None:
            expired_at = datetime.now(timezone.utc) + timedelta(minutes=5)

        operations.append(
            self.acp_contract_client.approve_allowance(
                amount.amount, amount.fare.contract_address
            )
        )

        if self._price_type == PriceType.PERCENTAGE and not skip_fee:
            fee_amount = int(self.price_value * 10000)
            fee_type = FeeType.PERCENTAGE_FEE
        else:
            fee_amount = (FareAmount(0, self.base_fare)).amount
            fee_type = FeeType.NO_FEE

        operations.append(
            self.acp_contract_client.create_payable_memo(
                job_id=self.id,
                content=content,
                amount_base_unit=amount.amount,
                recipient=self.client_address,
                fee_amount_base_unit=fee_amount,
                fee_type=fee_type,
                next_phase=ACPJobPhase.COMPLETED,
                memo_type=MemoType.PAYABLE_NOTIFICATION,
                expired_at=expired_at,
                token=amount.fare.contract_address,
            )
        )

        response = self.acp_contract_client.handle_operation(operations)
        return get_txn_hash_from_response(response)

    def perform_x402_payment(self, budget: float):
        try:
            payment_url = "/acp-budget"

            # Perform X402 request to get payment requirements
            x402_payable_requirements = self.acp_contract_client.perform_x402_request(
                payment_url, self.acp_contract_client.get_acp_version(), str(budget)
            )

            if not x402_payable_requirements.get("isPaymentRequired"):
                return

            accepts = x402_payable_requirements["data"].get("accepts", [])
            if not accepts:
                raise Exception("No X402 payment requirements found")

            requirement = accepts[0]

            x402_payment = self.acp_contract_client.generate_x402_payment(
                X402PayableRequest(
                    to=requirement["payTo"],
                    value=int(requirement["maxAmountRequired"]),
                    maxTimeoutSeconds=requirement["maxTimeoutSeconds"],
                    asset=requirement["asset"],
                ),
                X402PayableRequirements.model_validate(
                    x402_payable_requirements["data"]
                ),
            )
            encoded_payment = x402_payment.encodedPayment
            signature = x402_payment.signature
            message = x402_payment.message
            nonce = message.get("nonce") if message else None

            if not nonce:
                raise Exception("No nonce found in X402 message")

            self.acp_contract_client.update_job_x402_nonce(self.id, str(nonce))

            x402_response = self.acp_contract_client.perform_x402_request(
                payment_url,
                self.acp_contract_client.get_acp_version(),
                str(budget),
                encoded_payment,
            )
            if x402_response.get("isPaymentRequired"):
                # If payment is required, submit transfer with authorization and handle the operation
                operations = (
                    self.acp_contract_client.submit_transfer_with_authorization(
                        message["from"],
                        message["to"],
                        int(message["value"]),
                        int(message["validAfter"]),
                        int(message["validBefore"]),
                        message["nonce"],
                        signature,
                    )
                )
                self.acp_contract_client.handle_operation(operations)

            wait_ms = 2000
            max_wait_ms = 30000  # max 30 seconds of polling
            iteration_count = 0
            max_iterations = 10

            while True:
                x402_payment_details = (
                    self.acp_contract_client.get_x402_payment_details(self.id)
                )
                if x402_payment_details.is_budget_received:
                    break

                iteration_count += 1
                if iteration_count >= max_iterations:
                    raise Exception("X402 payment timed out")
                time.sleep(wait_ms / 1000.0)
                wait_ms = min(wait_ms * 2, max_wait_ms)

        except Exception as e:
            # Optionally: handle exception, log, or re-raise
            print(f"An error occurred during perform_x402_payment: {e}")
            raise

    def _deliver_cross_chain_payable(self, client_address: str, amount: FareAmountBase) -> str | None:
        if not amount.fare.chain_id:
            raise ACPError("Chain ID is required for cross chain payable delivery")

        chain_id = amount.fare.chain_id

        token_balance = getERC20Balance(
            self.acp_contract_client.public_clients[chain_id],
            amount.fare.contract_address,
            client_address,
        )

        if token_balance < amount.amount:
            raise ACPError("Insufficient token balance for cross chain payable delivery")

        asset_manager_address = self.acp_contract_client.get_asset_manager_address()
    
        token_allowance = getERC20Allowance(
            self.acp_contract_client.public_clients[chain_id],
            amount.fare.contract_address,
            client_address,
            asset_manager_address
        )

        approve_allowance_op = self.acp_contract_client.approve_allowance(
            amount.amount + token_allowance,
            amount.fare.contract_address,
            asset_manager_address
        )

        self.acp_contract_client.handle_operation([approve_allowance_op], chain_id)

        token_symbol = getERC20Symbol(
            self.acp_contract_client.public_clients[chain_id],
            amount.fare.contract_address,
        )

        create_memo_op = self.acp_contract_client.create_cross_chain_payable_memo(
            job_id=self.id,
            content=f"Performing cross chain transfer of {amount.amount} {token_symbol} to {client_address} on chain {chain_id}",
            token=amount.fare.contract_address,
            amount_base_unit=amount.amount,
            recipient=client_address,
            fee_amount_base_unit=0,
            fee_type=FeeType.NO_FEE,
            memo_type=MemoType.PAYABLE_TRANSFER,
            expired_at=datetime.now(timezone.utc) + timedelta(minutes=5),
            next_phase=ACPJobPhase.COMPLETED,
            destination_eid=get_destination_endpoint_id(chain_id),
            secured=True,
        )

        self.acp_contract_client.handle_operation([create_memo_op])

    def get_deliverable(self) -> Optional[DeliverablePayload]:
        deliverable = self.deliverable
        if not deliverable:
            return None

        if not isinstance(deliverable, str):
            return deliverable

        if not re.search(r"api/memo-contents/([0-9]+)$", deliverable):
            return deliverable

        content = self.acp_client.get_memo_content(deliverable)

        try:
            return json.loads(content)
        except (json.JSONDecodeError, TypeError):
            return content
    


from web3 import Web3

USDC_TOKEN_ADDRESS = {
    84532: "0x036CbD53842c5426634e7929541eC2318f3dCF7e",
    8453: "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
}

# EIP-712 "Authorization" type data specification for X402.
X402_AUTHORIZATION_TYPES = [
    {"name": "from", "type": "address"},
    {"name": "to", "type": "address"},
    {"name": "value", "type": "uint256"},
    {"name": "validAfter", "type": "uint256"},
    {"name": "validBefore", "type": "uint256"},
    {"name": "nonce", "type": "bytes32"},
]

HTTP_STATUS_CODES_X402 = {"Payment Required": 402, "OK": 200}

SINGLE_SIGNER_VALIDATION_MODULE_ADDRESS = Web3.to_checksum_address("0x00000000000099DE0BF6fA90dEB851E2A2df7d83")



from typing import Optional
from pydantic_settings import BaseSettings
from pydantic import field_validator


class EnvSettings(BaseSettings):
    WHITELISTED_WALLET_PRIVATE_KEY: Optional[str] = None
    BUYER_AGENT_WALLET_ADDRESS: Optional[str] = None
    SELLER_AGENT_WALLET_ADDRESS: Optional[str] = None
    EVALUATOR_AGENT_WALLET_ADDRESS: Optional[str] = None
    BUYER_GAME_TWITTER_ACCESS_TOKEN: Optional[str] = None
    SELLER_GAME_TWITTER_ACCESS_TOKEN: Optional[str] = None
    EVALUATOR_GAME_TWITTER_ACCESS_TOKEN: Optional[str] = None
    BUYER_ENTITY_ID: Optional[int] = None
    SELLER_ENTITY_ID: Optional[int] = None
    EVALUATOR_ENTITY_ID: Optional[int] = None

    @field_validator(
        "BUYER_AGENT_WALLET_ADDRESS",
        "SELLER_AGENT_WALLET_ADDRESS",
        "EVALUATOR_AGENT_WALLET_ADDRESS",
    )
    def validate_wallet_address(cls, v: str) -> str:
        if v is None:
            return None
        if not v.startswith("0x") or len(v) != 42:
            raise ValueError(
                "Wallet address must start with '0x' and be 42 characters long."
            )
        return v



import json
from datetime import datetime, timezone, timedelta
from typing import Any, Dict, Optional, Union, TYPE_CHECKING, List

from jsonschema import ValidationError, validate
from pydantic import BaseModel, field_validator, ConfigDict
from web3 import Web3
from web3.constants import ADDRESS_ZERO

from virtuals_acp.configs.configs import (
    BASE_SEPOLIA_CONFIG,
    BASE_MAINNET_CONFIG,
    BASE_SEPOLIA_ACP_X402_CONFIG,
    BASE_MAINNET_ACP_X402_CONFIG
)
from virtuals_acp.constants import USDC_TOKEN_ADDRESS
from virtuals_acp.contract_clients.base_contract_client import BaseAcpContractClient
from virtuals_acp.fare import FareAmount
from virtuals_acp.models import ACPJobPhase, MemoType, OperationPayload, PriceType

if TYPE_CHECKING:
    from virtuals_acp.client import VirtualsACP


class ACPJobOffering(BaseModel):
    acp_client: "VirtualsACP"
    contract_client: BaseAcpContractClient
    provider_address: str
    name: str
    price: float
    price_type: PriceType
    required_funds: bool
    sla_minutes: int
    requirement: Optional[Union[Dict[str, Any], str]] = None
    deliverable: Optional[Union[Dict[str, Any], str]] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("requirement", mode="before")
    def parse_requirement_schema(cls, v):
        if isinstance(v, str):
            try:
                return json.loads(json.dumps(v))
            except json.JSONDecodeError:
                return None
        return v

    def __str__(self):
        return f"ACPJobOffering({self.model_dump(exclude={'acp_client'})})"

    def __repr__(self) -> str:
        return self.__str__()

    def initiate_job(
        self,
        service_requirement: Union[Dict[str, Any], str],
        evaluator_address: Optional[str] = None,
    ) -> int:
        expired_at = datetime.now(timezone.utc) + timedelta(minutes=self.sla_minutes)

        # Validate against requirement schema if present
        if self.requirement:
            try:
                service_requirement = json.loads(json.dumps(service_requirement))
            except json.JSONDecodeError:
                raise ValueError(
                    f"Invalid JSON in service requirement. Required format: {json.dumps(self.requirement, indent=2)}"
                )

            if isinstance(self.requirement, dict):
                try:
                    validate(instance=service_requirement, schema=self.requirement)
                except ValidationError as e:
                    raise ValueError(f"Invalid service requirement: {str(e)}")

        final_service_requirement: Dict[str, Any] = {
            "name": self.name,
            "requirement": service_requirement,
            "priceValue": self.price,
            "priceType": self.price_type,
        }

        eval_addr = (
            Web3.to_checksum_address(evaluator_address)
            if evaluator_address
            else self.contract_client.agent_wallet_address
        )

        # Prepare fare amount based on this offering's price and contract's base fare
        fare_amount = FareAmount(
            self.price if self.price_type == PriceType.FIXED else 0,
            self.contract_client.config.base_fare,
        )

        # Lookup existing account between client and provider
        account = self.acp_client.get_by_client_and_provider(
            self.contract_client.agent_wallet_address,
            self.provider_address,
            self.contract_client,
        )

        base_contract_addresses = {
            BASE_SEPOLIA_CONFIG.contract_address.lower(),
            BASE_SEPOLIA_ACP_X402_CONFIG.contract_address.lower(),
            BASE_MAINNET_CONFIG.contract_address.lower(),
            BASE_MAINNET_ACP_X402_CONFIG.contract_address.lower(),
        }

        use_simple_create = (
            self.contract_client.config.contract_address.lower()
            in base_contract_addresses
        )

        chain_id = self.contract_client.config.chain_id
        usdc_token_address = USDC_TOKEN_ADDRESS[chain_id]
        is_usdc_payment_token = usdc_token_address == fare_amount.fare.contract_address
        
        # If the contract has x402_config and USDC is used, call create_job_with_x402
        is_x402_job =  bool(getattr(self.contract_client.config, "x402_config", None) and is_usdc_payment_token)
        if use_simple_create or not account:
            create_job_operation = self.contract_client.create_job(
                self.provider_address,
                eval_addr,
                expired_at,
                fare_amount.fare.contract_address,
                fare_amount.amount,
                "",
                is_x402_job=is_x402_job,
            )
        else:
            evaluator_address = (
                Web3.to_checksum_address(evaluator_address)
                if evaluator_address
                else ADDRESS_ZERO
            )
            create_job_operation = self.contract_client.create_job_with_account(
                account.id,
                evaluator_address or self.contract_client.agent_wallet_address,
                fare_amount.amount,
                fare_amount.fare.contract_address,
                expired_at,
                is_x402_job=is_x402_job,
            )

        response = self.contract_client.handle_operation([create_job_operation])

        job_id = self.contract_client.get_job_id(
            response,
            self.contract_client.agent_wallet_address,
            self.provider_address,
        )

        operations: List[OperationPayload] = []

        operation = self.contract_client.set_budget_with_payment_token(
            job_id,
            fare_amount.amount,
            fare_amount.fare.contract_address,
        )

        if operation:
            operations.append(operation)

        operations.append(
            self.contract_client.create_memo(
                job_id,
                json.dumps(final_service_requirement),
                MemoType.MESSAGE,
                True,
                ACPJobPhase.NEGOTIATION,
            )
        )
        self.contract_client.handle_operation(operations)

        return job_id


class ACPResourceOffering(BaseModel):
    acp_client: "VirtualsACP"
    name: str
    description: str
    url: str
    parameters: Optional[Dict[str, Any]]
    id: int

    model_config = ConfigDict(arbitrary_types_allowed=True)



import functools
import json
import warnings
from typing import Optional, Type, Union, Dict, Any
import base64

from pydantic import ValidationError

from virtuals_acp.models import T


def get_txn_hash_from_response(response: Dict[str, Any]) -> Optional[str]:
    try:
        return response.get("receipts", [])[0].get("transactionHash")
    except (IndexError, AttributeError):
        print(f"Error getting transaction hash from response: {response}")
        return None


def try_parse_json_model(content: str, model: Type[T]) -> Optional[T]:
    try:
        return model.model_validate_json(content)
    except (json.JSONDecodeError, ValidationError):
        return None


def try_validate_model(data: dict, model: Type[T]) -> Optional[T]:
    try:
        return model.model_validate(data)
    except ValidationError:
        return None


def prepare_payload(payload: Union[str, Dict[str, Any]]) -> str:
    return payload if isinstance(payload, str) else json.dumps(payload)


def deprecated(reason: str = "This function is deprecated and should not be used."):
    """Decorator to mark functions or methods as deprecated."""

    def decorator(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            warnings.warn(
                f"Call to deprecated function {func.__name__}: {reason}",
                category=DeprecationWarning,
                stacklevel=2,
            )
            return func(*args, **kwargs)

        return wrapped

    return decorator


def safe_base64_encode(data: Union[str, bytes]) -> str:
    if isinstance(data, str):
        data_bytes = data.encode("utf-8")
    else:
        data_bytes = data
    return base64.b64encode(data_bytes).decode("utf-8")

def get_destination_endpoint_id(chain_id: int) -> int:
    id_to_eid = {
        84532: 40245,    # baseSepolia.id
        11155111: 40161, # sepolia.id
        80002: 40267,    # polygonAmoy.id
        421614: 40231,   # arbitrumSepolia.id
        97: 40102,       # bscTestnet.id
        8453: 30184,     # base.id
        1: 30101,        # mainnet.id
        137: 30109,      # polygon.id
        42161: 30110,    # arbitrum.id
        56: 30102,       # bsc.id
    }
    if chain_id in id_to_eid:
        return id_to_eid[chain_id]
    raise ValueError(f"Unsupported chain ID: {chain_id}")
    
def get_destination_chain_id(endpoint_id: int) -> int:
    eid_to_id = {
        40245: 84532,     # baseSepolia.id
        40161: 11155111,  # sepolia.id
        40267: 80002,     # polygonAmoy.id
        40231: 421614,    # arbitrumSepolia.id
        40102: 97,        # bscTestnet.id
        30184: 8453,      # base.id
        30101: 1,         # mainnet.id
        30109: 137,       # polygon.id
        30110: 42161,     # arbitrum.id
        30102: 56,        # bsc.id
    }
    if endpoint_id in eid_to_id:
        return eid_to_id[endpoint_id]
    raise ValueError(f"Unsupported endpoint ID: {endpoint_id}")


import json
from abc import ABC, abstractmethod
from datetime import datetime
from decimal import Decimal
import math
from typing import Dict, Any, Optional, List, cast

from eth_typing import ABIEvent
from ens.utils import is_none_or_zero_address
from web3 import Web3
from web3.contract import Contract
from eth_utils.abi import event_abi_to_log_topic

from virtuals_acp.abis.erc20_abi import ERC20_ABI
from virtuals_acp.abis.flat_token_v2_abi import FIAT_TOKEN_V2_ABI
from virtuals_acp.abis.single_signer_validation_module_abi import SINGLE_SIGNER_VALIDATION_MODULE_ABI
from virtuals_acp.abis.weth_abi import WETH_ABI
from virtuals_acp.constants import SINGLE_SIGNER_VALIDATION_MODULE_ADDRESS
from virtuals_acp.fare import WETH_FARE
from virtuals_acp.configs.configs import ACPContractConfig
from virtuals_acp.exceptions import ACPError
from virtuals_acp.models import (
    ACPJobPhase,
    MemoType,
    FeeType,
    AcpJobX402PaymentDetails,
    X402PayableRequest,
    X402Payment,
    X402PayableRequirements,
    OperationPayload,
    OffChainJob,
)

#  TODO: This function is not used anywhere, should we add it to Python SDK?
# createMemoWithMetadata
# signTypedData
# signMessage
# sendTransaction


class BaseAcpContractClient(ABC):
    def __init__(self, agent_wallet_address: str, config: ACPContractConfig):
        self.agent_wallet_address = Web3.to_checksum_address(agent_wallet_address)
        self.config = config
        self.w3 = Web3(Web3.HTTPProvider(config.rpc_url))
        self.public_clients: Dict[int, Web3] = {}

        self.chain = config.chain
        self.abi = config.abi
        self.contract_address = Web3.to_checksum_address(config.contract_address)

        if not self.w3.is_connected():
            raise ConnectionError(f"Failed to connect to RPC: {config.rpc_url}")

        self.contract: Contract = self.w3.eth.contract(
            address=Web3.to_checksum_address(config.contract_address),
            abi=self.abi,
        )
        self.token_contract: Contract = self.w3.eth.contract(
            address=Web3.to_checksum_address(config.base_fare.contract_address),
            abi=self.abi,
        )

        job_created_event_abi = next(
            (
                item
                for item in config.abi
                if item.get("type") == "event" and item.get("name") == "JobCreated"
            ),
            None,
        )

        if not job_created_event_abi:
            raise ACPError("JobCreated event not found in ACP_ABI")

        self.job_created_event_signature_hex = (
            "0x" + event_abi_to_log_topic(cast(ABIEvent, job_created_event_abi)).hex()
        )

        agent_smart_contract_code = self.w3.eth.get_code(Web3.to_checksum_address(self.agent_wallet_address))
        is_account_deployed = len(agent_smart_contract_code) > 0
        if not is_account_deployed:
            raise ACPError(f"ACP Contract Client validation failed: agent account {self.agent_wallet_address} is not deployed on-chain")

    def validate_session_key_on_chain(
        self,
        session_signer_address: str,
        session_entity_key_id: int
    ) -> None:
        single_signer_validation_contract: Contract = self.w3.eth.contract(
            address=SINGLE_SIGNER_VALIDATION_MODULE_ADDRESS,
            abi=SINGLE_SIGNER_VALIDATION_MODULE_ABI,
        )
        on_chain_signer_address = (
            single_signer_validation_contract
                .functions
                .signers(session_entity_key_id, self.agent_wallet_address)
                .call()
        )

        if is_none_or_zero_address(on_chain_signer_address):
            raise ACPError(
                "ACP Contract Client validation failed:\n" +
                json.dumps(
                    {
                        "reason": "no whitelisted wallet registered on-chain for entity id",
                        "entity_id": session_entity_key_id,
                        "agent_wallet_address": self.agent_wallet_address,
                    },
                    indent=2
                )
            )

        if on_chain_signer_address.lower() != session_signer_address.lower():
            raise ACPError(
                "ACP Contract Client validation failed:\n" +
                json.dumps(
                    {
                        "agent_wallet_address": self.agent_wallet_address,
                        "entity_id": session_entity_key_id,
                        "given_whitelisted_wallet_address": session_signer_address,
                        "expected_whitelisted_wallet_address": on_chain_signer_address,
                        "reason": "session signer address mismatch",
                    },
                    indent=2
                )
            )

    @abstractmethod
    def get_acp_version(self) -> str:
        pass

    @abstractmethod
    def get_asset_manager_address(self) -> str:
        pass

    @abstractmethod
    def sign_typed_data(self, typed_data: dict[str, Any]) -> str:
        pass

    def _build_user_operation(
        self,
        method_name: str,
        args: List[Any],
        contract_address: Optional[str] = None,
        abi: Optional[List[Dict[str, Any]]] = None,
    ) -> Dict[str, Any]:
        """
        Build a single-call user operation to invoke a contract method.
        If no ABI is provided, defaults to the ACP contract ABI.
        """
        target_abi = abi or self.abi
        target_address = Web3.to_checksum_address(
            contract_address or self.config.contract_address
        )

        target_contract = self.w3.eth.contract(address=target_address, abi=target_abi)
        encoded_data = target_contract.encode_abi(method_name, args=args)

        return {"to": target_address, "data": encoded_data}

    @abstractmethod
    def handle_operation(self, trx_data: List[OperationPayload], chain_id: Optional[int] = None) -> Dict[str, Any]:
        pass

    @abstractmethod
    def get_job_id(
        self, response: Dict[str, Any], client_address: str, provider_address: str
    ) -> int:
        """Abstract method to retrieve a job ID from a transaction hash and related addresses."""
        pass

    def _format_amount(self, amount: float) -> int:
        return int(Decimal(str(amount)) * (10**self.config.base_fare.decimals))

    def update_account_metadata(
        self, account_id: int, metadata: str
    ) -> OperationPayload:
        operation = self._build_user_operation(
            "updateAccountMetadata",
            [account_id, metadata],
            self.config.contract_address,
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def create_job(
        self,
        provider_address: str,
        evaluator_address: str,
        expired_at: datetime,
        payment_token_address: str,
        budget_base_unit: int,
        metadata: str,
        is_x402_job: bool = False
    ) -> OperationPayload:
        fn_name = "createX402Job" if is_x402_job else "createJob"
        operation = self._build_user_operation(
            fn_name,
            [
                Web3.to_checksum_address(provider_address),
                Web3.to_checksum_address(evaluator_address),
                math.floor(expired_at.timestamp()),
                payment_token_address,
                budget_base_unit,
                metadata,
            ],
            self.config.contract_address,
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def create_job_with_account(
        self,
        account_id: int,
        evaluator_address: str,
        budget_base_unit: int,
        payment_token_address: str,
        expired_at: datetime,
        is_x402_job: bool = False
    ) -> OperationPayload:
        fn_name = "createX402JobWithAccount" if is_x402_job else "createJobWithAccount"
        operation = self._build_user_operation(
            fn_name,
            [
                account_id,
                Web3.to_checksum_address(evaluator_address),
                budget_base_unit,
                Web3.to_checksum_address(payment_token_address),
                math.floor(expired_at.timestamp()),
            ],
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def approve_allowance(
        self,
        amount_base_unit: int,
        payment_token_address: Optional[str] = None,
        spender_address: Optional[str] = None,
    ) -> OperationPayload:
        operation = self._build_user_operation(
            "approve",
            [spender_address or self.config.contract_address, amount_base_unit],
            contract_address=payment_token_address,
            abi=ERC20_ABI,
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def create_payable_memo(
        self,
        job_id: int,
        content: str,
        amount_base_unit: int,
        recipient: str,
        fee_amount_base_unit: int,
        fee_type: FeeType,
        next_phase: ACPJobPhase,
        memo_type: MemoType,
        expired_at: datetime,
        token: Optional[str] = None,
        secured: bool = True,
    ) -> OperationPayload:
        operation = self._build_user_operation(
            "createPayableMemo",
            [
                job_id,
                content,
                token or self.config.base_fare.contract_address,
                amount_base_unit,
                Web3.to_checksum_address(recipient),
                fee_amount_base_unit,
                fee_type,
                memo_type,
                math.floor(expired_at.timestamp()),
                secured,
                next_phase,
            ],
            self.config.contract_address,
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def create_cross_chain_payable_memo(
        self,
        job_id: int,
        content: str,
        token: str,
        amount_base_unit: int,
        recipient: str,
        fee_amount_base_unit: int,
        fee_type: FeeType,
        memo_type: MemoType,
        expired_at: datetime,
        next_phase: ACPJobPhase,
        destination_eid: int,
        secured: bool = True,
    ) -> OperationPayload:
        operation = self._build_user_operation(
            "createCrossChainPayableMemo",
            [
                job_id,
                content,
                token,
                amount_base_unit,
                Web3.to_checksum_address(recipient),
                fee_amount_base_unit,
                fee_type,
                memo_type,
                math.floor(expired_at.timestamp()),
                secured,
                next_phase,
                destination_eid,
            ],
            self.config.contract_address,
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def create_memo(
        self,
        job_id: int,
        content: str,
        memo_type: MemoType,
        is_secured: bool,
        next_phase: ACPJobPhase,
    ) -> OperationPayload:
        operation = self._build_user_operation(
            "createMemo",
            [job_id, content, memo_type.value, is_secured, next_phase.value],
            self.config.contract_address,
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def sign_memo(
        self, memo_id: int, is_approved: bool, reason: Optional[str] = ""
    ) -> OperationPayload:
        operation = self._build_user_operation(
            "signMemo", [memo_id, is_approved, reason], self.config.contract_address
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def set_budget_with_payment_token(
        self,
        job_id: int,
        budget_base_unit: int,
        payment_token_address: Optional[str] = None,
    ) -> OperationPayload | None:
        return None

    def wrap_eth(self, amount_base_unit: int) -> OperationPayload:
        weth_contract = self.w3.eth.contract(
            address=Web3.to_checksum_address(WETH_FARE.contract_address),
            abi=WETH_ABI,
        )
        # Build a user operation (single call)
        trx_data = self._build_user_operation(
            method_name="deposit",
            args=[],
            contract_address=weth_contract.address,
            abi=WETH_ABI,
        )

        operation = OperationPayload(
            data=trx_data["data"],
            to=trx_data["to"],
            value=amount_base_unit,
        )

        return operation

    def get_x402_payment_details(self, job_id: int) -> AcpJobX402PaymentDetails:
        """Get X402 payment details for a job."""
        try:
            x402_config = self.config.x402_config
            if not x402_config or not getattr(x402_config, "url", None):
                return AcpJobX402PaymentDetails(is_x402=False, is_budget_received=False)

            # Call the contract function
            result = self.contract.functions.x402PaymentDetails(job_id).call()

            return AcpJobX402PaymentDetails(
                is_x402=result[0], is_budget_received=result[1]
            )
        except Exception as e:
            raise ACPError("Failed to get X402 payment details", e)

    @abstractmethod
    def update_job_x402_nonce(self, job_id: int, nonce: str) -> OffChainJob:
        """Abstract method to update the X402 nonce for a job."""
        pass

    @abstractmethod
    def generate_x402_payment(
        self, payable_request: X402PayableRequest, requirements: X402PayableRequirements
    ) -> X402Payment:
        """Abstract method to generate an X402 payment."""
        pass

    @abstractmethod
    def perform_x402_request(
        self, url: str, version: str, budget: Optional[str] = None, signature: Optional[str] = None
    ) -> Dict[str, Any]:
        """Abstract method to perform an X402 request."""
        pass

    def submit_transfer_with_authorization(
        self,
        from_address: str,
        to_address: str,
        value: int,
        valid_after: int,
        valid_before: int,
        nonce: str,
        signature: str,
    ) -> List[OperationPayload]:
        try:
            contract = self.w3.eth.contract(
                address=self.config.base_fare.contract_address, abi=FIAT_TOKEN_V2_ABI
            )

            data = contract.encode_abi(
                "transferWithAuthorization",
                args=[
                    from_address,
                    to_address,
                    value,
                    valid_after,
                    valid_before,
                    nonce,
                    signature,
                ],
            )

            payload = OperationPayload(
                data=data, to=self.config.base_fare.contract_address
            )
            return [payload]

        except Exception as error:
            from virtuals_acp.exceptions import ACPError

            raise ACPError("Failed to submit TransferWithAuthorization", error)



import secrets
import time
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Any, List, Optional, Union

import requests
from eth_account import Account
from eth_account.messages import encode_defunct
from eth_account.messages import encode_typed_data
from eth_utils.conversions import to_hex

from virtuals_acp.configs.configs import BASE_SEPOLIA_CONFIG, ACPContractConfig
from virtuals_acp.models import ChainConfig, OperationPayload

MAX_RETRIES = 10


class PermissionType(str, Enum):
    ROOT = "root"


class SignatureRequestType(str, Enum):
    PERSONAL_SIGN = "personal_sign"
    ETH_SIGN_TYPED_DATA_V4 = "eth_signTypedData_v4"


@dataclass
class SignatureRequest:
    type: SignatureRequestType
    data: Union[str, Dict[str, Any]]


@dataclass
class KeyInfo:
    public_key: str
    type: str = "secp256k1"


@dataclass
class Permission:
    type: PermissionType


class AlchemyRPCClient:
    def __init__(self, base_url: str = BASE_SEPOLIA_CONFIG.alchemy_base_url):
        self.base_url = base_url
        self.session = requests.Session()
        self.session.headers.update({"Content-Type": "application/json"})

    def request(self, method: str, params: List[Any]) -> Any:
        try:
            """Make a JSON-RPC request to the Alchemy API"""
            payload = {"jsonrpc": "2.0", "method": method, "params": params, "id": 1}
            response = self.session.post(self.base_url, json=payload)
            result = response.json()
            if result.get("error"):
                raise Exception(f"RPC Error: {result['error']}")

            if result.get("result"):
                return result.get("result")

            return result
        except Exception as e:
            print(f"Error on request: {e}")
            raise e

    # Typed wallet method wrappers for better API
    def wallet_request_account(self, signer_address: str) -> Dict[str, Any]:
        """Request an account from Alchemy"""
        return self.request(
            "wallet_requestAccount", [{"signerAddress": signer_address}]
        )

    def wallet_create_account(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Create a new account"""
        return self.request("wallet_createAccount", [params])

    def wallet_prepare_calls(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Prepare calls for execution"""
        return self.request("wallet_prepareCalls", [params])

    def wallet_send_prepared_calls(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Send prepared calls"""
        return self.request("wallet_sendPreparedCalls", [params])

    def wallet_create_session(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Create a session"""
        return self.request("wallet_createSession", [params])

    def wallet_get_calls_status(self, params: str) -> Dict[str, Any]:
        """Get calls status"""
        return self.request("wallet_getCallsStatus", [params])


class AlchemyAccountKit:
    def __init__(
        self,
        config: ACPContractConfig,
        agent_wallet_address: str,
        entity_id: int,
        owner_account: Account,
        chain_id: Optional[int] = None,
        chains: Optional[List[ChainConfig]] = None,
    ):
        """
        Initialize the Alchemy Account Kit

        Args:
            agent_wallet_address: Agent wallet address
            entity_id: Entity ID
            chain_id: Chain ID to use (defaults to BASE_SEPOLIA_CONFIG.chain_id)
        """
        self.chain_id = chain_id or config.chain_id
        self.rpc_client = AlchemyRPCClient(config.alchemy_base_url)
        self.account_address = agent_wallet_address
        self.owner_account = owner_account
        self.rpc_clients = {}

        if chains:
            for chain in chains:
                self.rpc_clients[chain.chain_id] = AlchemyRPCClient(f"{config.alchemy_base_url}?chainId={chain.chain_id}")

        self.permissions_context = self.create_session(entity_id)

    def sign_signature_request(
        self, request: SignatureRequest, account: Account
    ) -> str:
        """Handle signature requests of both types"""
        if request.type == SignatureRequestType.PERSONAL_SIGN:
            # For personal_sign, the data is raw hex data
            if isinstance(request.data, dict):
                # Sign the message using the hex data directly
                raw_data = request.data.get("raw")
                signed_message = account.sign_message(
                    encode_defunct(hexstr=str(raw_data))
                )
                return signed_message.signature.hex()
            else:
                raise ValueError("Personal sign data must be a hex string")

        elif request.type == SignatureRequestType.ETH_SIGN_TYPED_DATA_V4:
            # For typed data, the data should be a dictionary with the structured data
            if isinstance(request.data, dict):
                # Create the structured data message
                structured_data = encode_typed_data(request.data)
                signed_message = account.sign_message(structured_data)
                return signed_message.signature.hex()
            else:
                raise ValueError("Typed data must be a dictionary")

        else:
            raise ValueError(f"Unsupported signature request type: {request.type}")

    def create_account(self, params: Dict[str, Any]) -> Dict[str, Any]:
        result = self.rpc_client.wallet_create_account(params)

        return result

    def create_session(self, entity_id: int) -> str:
        # Prepare permissions context
        permissions_context_version = "0x02"  # REMOTE_MODE_PERMISSIONS_CONTEXT
        is_global_validation = "01"  # Should be int, not string

        # Concatenate hex values (equivalent to concatHex in viem)
        # Concatenate hex values by removing 0x prefix from subsequent values
        return (
            permissions_context_version
            + is_global_validation  # Remove 0x prefix
            + to_hex(entity_id)[2:].zfill(8)  # Remove 0x prefix and pad to 8 chars
        )

    def get_random_nonce(self, bits=152):
        bytes_length = bits // 8
        random_bytes = secrets.token_bytes(bytes_length)
        hex_str = random_bytes.hex()
        return "0x" + hex_str

    def prepare_calls(
        self,
        calls: List[OperationPayload],
        capabilities: Optional[Dict[str, Any]] = None,
        chain_id: Optional[int] = None,
    ) -> Dict[str, Any]:
        if not self.account_address:
            raise ValueError("Must request account first")
        if not self.permissions_context:
            raise ValueError("Must create session first")

        final_capabilities = {
            "permissions": {"context": self.permissions_context},
            "paymasterService": {"policyId": BASE_SEPOLIA_CONFIG.alchemy_policy_id},
            "nonceOverride": {"nonceKey": self.get_random_nonce()},
        }

        if capabilities:
            final_capabilities.update(capabilities)

        params = {
            "from": self.account_address,
            "chainId": to_hex(chain_id or self.chain_id),
            "calls": [call.model_dump(exclude_none=True) for call in calls],
            "capabilities": final_capabilities,
        }

        if chain_id:
            return self.rpc_clients[chain_id].wallet_prepare_calls(params)
        else:
            return self.rpc_client.wallet_prepare_calls(params)

    def send_prepared_calls(
        self, prepare_calls_result: Dict[str, Any], chain_id: Optional[int] = None
    ) -> Dict[str, Any]:
        if not self.permissions_context:
            raise ValueError("Must create session first")

        # Sign the prepare calls result using the session key
        prepare_calls_signature_request_data = prepare_calls_result["signatureRequest"]
        prepare_calls_signature_request = SignatureRequest(
            type=SignatureRequestType(prepare_calls_signature_request_data["type"]),
            data=prepare_calls_signature_request_data["data"],
        )

        user_op_signature = self.sign_signature_request(
            prepare_calls_signature_request, self.owner_account
        )

        # Prepare the parameters for sending prepared calls
        send_prepared_calls_params = prepare_calls_result.copy()
        # Remove the signatureRequest from the params
        send_prepared_calls_params.pop("signatureRequest", None)

        send_prepared_calls_params["signature"] = {
            "type": "secp256k1",
            "data": "0x" + user_op_signature,
        }

        send_prepared_calls_params["capabilities"] = {
            "permissions": {"context": self.permissions_context}
        }

        if chain_id:
            return self.rpc_clients[chain_id].wallet_send_prepared_calls(send_prepared_calls_params)
        else:
            return self.rpc_client.wallet_send_prepared_calls(send_prepared_calls_params)

    def wait_for_call_status(self, prepared_call_id: str) -> Dict[str, Any]:
        retries = MAX_RETRIES
        while True:
            try:
                status = self.rpc_client.wallet_get_calls_status(prepared_call_id)

                if status["status"] == 200:
                    return status

                raise Exception("Retrying...")
            except Exception as e:
                retries -= 1

                if retries == 0:
                    raise Exception("Failed to get call status")

            time.sleep(0.1 * (MAX_RETRIES - retries))

    def handle_user_operation(
        self, calls: List[OperationPayload], capabilities=None, chain_id: Optional[int] = None
    ) -> Dict[str, Any]:
        if capabilities is None:
            capabilities = {}

        retries = MAX_RETRIES

        # Increase the max fee per gas by 10%
        additional_capabilities = {"maxFeePerGas": {"multiplier": 1.1}}
        capabilities.update(additional_capabilities)

        while True:
            try:
                prepare_result = self.prepare_calls(calls, capabilities, chain_id)
                result = self.send_prepared_calls(prepare_result, chain_id)
                return self.wait_for_call_status(result["preparedCallIds"][0])
            except Exception as e:
                retries -= 1

                if retries == 0:
                    raise Exception(e)

            time.sleep(0.1 * (MAX_RETRIES - retries))



from abc import ABC, abstractmethod
from decimal import Decimal, ROUND_DOWN
from typing import Optional, Union
from web3 import Web3
from web3.contract import Contract

from virtuals_acp.exceptions import ACPError

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from virtuals_acp.configs.configs import ACPContractConfig


class Fare:
    def __init__(self, contract_address: str, decimals: int, chain_id: Optional[int] = None):
        self.contract_address = Web3.to_checksum_address(contract_address)
        self.decimals = decimals
        self.chain_id = chain_id

    def format_amount(self, amount: Union[int, float, Decimal]) -> int:
        """Convert to smallest unit (like parseUnits)."""
        amount_decimal = Decimal(str(amount)).scaleb(self.decimals)
        return int(amount_decimal.to_integral_value(rounding=ROUND_DOWN))

    @staticmethod
    def from_contract_address(
        contract_address: str, config: "ACPContractConfig", chain_id: Optional[int] = None
    ) -> "Fare":
        if Web3.to_checksum_address(contract_address) == Web3.to_checksum_address(
            config.base_fare.contract_address
        ):
            return config.base_fare

        rpc_url = config.rpc_url

        if chain_id and chain_id != config.chain_id:
            rpc_url = next(c.rpc_url for c in config.chains if c.chain_id == chain_id)

        if not rpc_url:
            raise ACPError(f"No RPC URL found for chain ID: {chain_id}")
        
        w3 = Web3(Web3.HTTPProvider(rpc_url))

        erc20_abi = [
            {
                "constant": True,
                "inputs": [],
                "name": "decimals",
                "outputs": [{"name": "", "type": "uint8"}],
                "type": "function",
            }
        ]

        contract: Contract = w3.eth.contract(
            address=Web3.to_checksum_address(contract_address), abi=erc20_abi
        )
        decimals = contract.functions.decimals().call()
        return Fare(contract_address, decimals, chain_id)


class FareAmountBase(ABC):
    def __init__(self, amount: int, fare: Fare):
        self.amount = amount
        self.fare = fare

    def __repr__(self) -> str:
        return (
            f"<FareAmount amount={self.amount} "
            f"token={self.fare.contract_address} "
            f"decimals={self.fare.decimals}>"
        )

    __str__ = __repr__

    @abstractmethod
    def add(self, other: "FareAmountBase") -> "FareAmountBase":
        pass

    @staticmethod
    def from_contract_address(
        amount: Union[int, float], contract_address: str, config: "ACPContractConfig"
    ) -> "FareAmountBase":
        fare = Fare.from_contract_address(contract_address, config)
        if isinstance(amount, float):
            return FareAmount(amount, fare)
        return FareBigInt(int(amount), fare)


class FareAmount(FareAmountBase):
    def __init__(self, fare_amount: Union[int, float], fare: Fare):
        def truncate_to_6_decimals(value: str) -> str:
            d = Decimal(value)
            return str(d.quantize(Decimal("0.000001"), rounding=ROUND_DOWN))

        truncated = truncate_to_6_decimals(str(fare_amount))
        super().__init__(fare.format_amount(Decimal(truncated)), fare)

    def add(self, other: FareAmountBase) -> "FareAmountBase":
        if self.fare.contract_address != other.fare.contract_address:
            raise ACPError("Token addresses do not match")
        return FareBigInt(self.amount + other.amount, self.fare)


class FareBigInt(FareAmountBase):
    def __init__(self, amount: int, fare: Fare):
        super().__init__(amount, fare)

    def add(self, other: FareAmountBase) -> "FareAmountBase":
        if self.fare.contract_address != other.fare.contract_address:
            raise ACPError("Token addresses do not match")
        return FareBigInt(self.amount + other.amount, self.fare)


# --- Declared Fare Instances ---
WETH_FARE = Fare("0x4200000000000000000000000000000000000006", 18)
ETH_FARE = Fare("0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", 18)



# virtuals_acp/client.py

import json
import logging
import signal
import sys
import threading
import jwt
import socketio
import requests
import time

from datetime import datetime, timezone, timedelta
from importlib.metadata import version
from typing import List, Optional, Union, Dict, Any, Callable
from web3 import Web3
from requests.auth import AuthBase

from virtuals_acp.account import ACPAccount
from virtuals_acp.configs.configs import (
    BASE_MAINNET_ACP_X402_CONFIG,
    BASE_SEPOLIA_ACP_X402_CONFIG,
    BASE_SEPOLIA_CONFIG,
    BASE_MAINNET_CONFIG,
)
from virtuals_acp.constants import USDC_TOKEN_ADDRESS
from virtuals_acp.contract_clients.base_contract_client import BaseAcpContractClient
from virtuals_acp.exceptions import ACPApiError, ACPError
from virtuals_acp.fare import FareAmountBase
from virtuals_acp.job import ACPJob
from virtuals_acp.job_offering import ACPJobOffering, ACPResourceOffering
from virtuals_acp.memo import ACPMemo
from virtuals_acp.models import (
    ACPAgentSort,
    ACPJobPhase,
    ACPGraduationStatus,
    ACPMemoState,
    ACPOnlineStatus,
    MemoType,
    IACPAgent,
    ACPMemoStatus,
    PriceType,
)

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
)
logger = logging.getLogger("ACPClient")



class BearerAuth(AuthBase):
    def __init__(self, get_access_token: Callable[[], str]):
        self._get_access_token = get_access_token
        self._access_token: Optional[str] = None

    def __call__(self, req: requests.PreparedRequest):
        if not self._access_token:
            self._access_token = self._get_access_token()
        req.headers["authorization"] = f"Bearer {self._access_token}"
        return req

    def clear_token(self):
        self._access_token = None


class ACPApiClient:
    def __init__(self, acp_contract_client: BaseAcpContractClient, acp_url: str, wallet_address: str, require_auth: bool = False):
        self.acp_contract_client = acp_contract_client
        self.base_url = acp_url
        self.wallet_address = wallet_address
        self.require_auth = require_auth
        self.session = requests.Session()
        
        self.access_token: Optional[str] = None
        self.auth: Optional[BearerAuth] = None
        if require_auth:
            self.auth = BearerAuth(self.get_access_token)
            self.session.auth = self.auth
            self.session.headers["wallet-address"] = wallet_address
            

    def request(
        self,
        method: str,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Dict[str, Any]] = None,
        err_callback: Optional[Callable[[requests.RequestException], None]] = None,
    ) -> Optional[Any]:
        if self.base_url in path:
            # absolute URL, use as is
            url = path
        else:
            url = f"{self.base_url}/{path}"
        
        try:
            resp = self.session.request(method, url, params=params, json=data)

            if resp.status_code == 401 and self.require_auth and self.auth:
                self.auth.clear_token()
                resp = self.session.request(method, url, params=params, json=data)

            resp.raise_for_status()
            return resp.json().get("data")
        except requests.RequestException as err:
            if err_callback:
                err_callback(err)
                return None

            if hasattr(err, "response") and err.response is not None:
                try:
                    error_message = err.response.json().get("error", {}).get("message")
                    if error_message:
                        raise ACPApiError(error_message) from err
                except (ValueError, AttributeError, KeyError):
                    pass

            raise ACPApiError(f"Failed to fetch {path}: {err}") from err
        except Exception as err:
            raise ACPApiError(
                f"Failed to fetch ACP Endpoint: {path} (network error)"
            ) from err

    def get_access_token(self) -> str:
        needs_refresh = self.access_token is None

        if self.access_token:
            decoded = jwt.decode(self.access_token, options={"verify_signature": False})
            if decoded.get("exp") and decoded["exp"] - 300 < time.time():
                needs_refresh = True

        if not needs_refresh:
            # Access token is still valid
            if self.access_token:
                return self.access_token
            else:
                raise Exception("Access token needs refreshing!")

        self.access_token = self.refresh_token()
        return self.access_token

    def refresh_token(self) -> str:
        challenge = self.get_auth_challenge()
        signature = self.acp_contract_client.sign_typed_data(challenge)

        verified = self.verify_auth_challenge(
            wallet_address=challenge["message"]["walletAddress"],
            nonce=challenge["message"]["nonce"],
            expires_at=challenge["message"]["expiresAt"],
            signature=signature,
        )

        return verified["accessToken"]

    def get_auth_challenge(self):
        try:
            response = requests.get(
                f"{self.base_url}/auth/challenge",
                params={"walletAddress": self.wallet_address},
            )
            response.raise_for_status()
            return response.json()["data"]
        except requests.RequestException as err:
            error_data = err.response.json() if err.response is not None else None
            print(f"Failed to get auth challenge: {error_data}")
            raise Exception("Failed to get auth challenge") from err

    def verify_auth_challenge(self, wallet_address: str, nonce: str, expires_at: int, signature: str):
        try:
            response = requests.post(
                f"{self.base_url}/auth/verify-typed-signature",
                json={
                    "walletAddress": wallet_address,
                    "nonce": nonce,
                    "expiresAt": expires_at,
                    "signature": signature,
                },
            )
            response.raise_for_status()
            return response.json()["data"]
        except requests.RequestException as err:
            raise Exception("Failed to verify auth challenge") from err


class VirtualsACP:
    def __init__(
        self,
        acp_contract_clients: Union[BaseAcpContractClient, List[BaseAcpContractClient]],
        on_new_task: Optional[Callable] = None,
        on_evaluate: Optional[Callable] = None,
        custom_rpc_url: Optional[str] = None,
        skip_socket_connection: Optional[bool] = False,
    ):
        # Handle both single client and list of clients
        if isinstance(acp_contract_clients, list):
            self.contract_clients = acp_contract_clients
        else:
            self.contract_clients = [acp_contract_clients]

        if len(self.contract_clients) == 0:
            raise ACPError("ACP contract client is required")

        # Validate all clients have the same agent wallet address
        first_agent_address = self.contract_clients[0].agent_wallet_address
        for client in self.contract_clients:
            if client.agent_wallet_address != first_agent_address:
                raise ACPError(
                    "All contract clients must have the same agent wallet address"
                )

        self.acp_client = ACPApiClient(self.acp_contract_client, self.acp_url, self.wallet_address, require_auth=True)
        self.no_auth_acp_client = ACPApiClient(self.acp_contract_client, self.acp_url, self.wallet_address)

        # Socket.IO setup
        self.on_new_task = on_new_task
        self.on_evaluate = on_evaluate or self._default_on_evaluate

        if not skip_socket_connection:
            self.sio = socketio.Client()
            self.init()

    @property
    def acp_contract_client(self):
        """Get the first contract client (for backward compatibility)."""
        return self.contract_clients[0]

    @property
    def wallet_address(self):
        """Get the wallet address from the first contract client."""
        return Web3.to_checksum_address(self.acp_contract_client.agent_wallet_address)

    @property
    def acp_url(self):
        """Get the ACP URL from the first contract client."""
        return self.acp_contract_client.config.acp_api_url

    def init(self):
        logger.info(f"Initializing socket")
        
        try:
            auth_data = {
                "walletAddress": self.wallet_address,
                "accessToken": self.acp_client.get_access_token()
            }
            headers_data = {
                "x-sdk-version": version("virtuals_acp"),
                "x-sdk-language": "python",
                "x-contract-address": self.contract_clients[0].contract_address,
            }

            self.sio.connect(
                url=self.acp_url,
                auth=auth_data,
                headers=headers_data,
                transports=["websocket"],
                retry=True,
            )

            def cleanup(sig, frame):
                self.sio.disconnect()
                sys.exit(0)

            self.sio.on("roomJoined", self._on_room_joined)
            self.sio.on("onEvaluate", self._on_evaluate)
            self.sio.on("onNewTask", self._on_new_task)

            signal.signal(signal.SIGINT, cleanup)
            signal.signal(signal.SIGTERM, cleanup)
        except Exception as e:
            logger.error(f"Failed to connect to socket server: {e}")
    
    def contract_client_by_address(self, address: Optional[str]):
        """Find contract client by contract address."""
        if not address:
            return self.contract_clients[0]

        for client in self.contract_clients:
            if (
                hasattr(client, "contract_address")
                and client.contract_address == address
            ):
                return client

        raise ACPError("ACP contract client not found")

    def _default_on_evaluate(self, job: ACPJob):
        """Default handler for job evaluation events."""
        job.evaluate(True, "Evaluated by default")

    def _on_room_joined(self, data):
        logger.info("Joined ACP Room", data)  # Send acknowledgment back to server
        return True

    def _on_evaluate(self, data):
        if self.on_evaluate:
            try:
                threading.Thread(target=self.handle_evaluate, args=(data,)).start()
                return True
            except Exception as e:
                logger.warning(f"Error in onEvaluate handler: {e}")
                return False

    def _on_new_task(self, data):
        if self.on_new_task:
            try:
                threading.Thread(target=self.handle_new_task, args=(data,)).start()
                return True
            except Exception as e:
                logger.warning(f"Error in onNewTask handler: {e}")
                return False

    def handle_new_task(self, data) -> None:
        memo_to_sign_id = data.get("memoToSign")

        memos = [
            ACPMemo(
                contract_client=self.contract_client_by_address(
                    data.get("contractAddress")
                ),
                id=memo.get("id"),
                type=MemoType(int(memo.get("memoType"))),
                content=memo.get("content"),
                next_phase=ACPJobPhase.from_value(memo.get("nextPhase")),
                status=ACPMemoStatus(memo.get("status")),
                signed_reason=memo.get("signedReason"),
                expiry=(
                    datetime.fromtimestamp(int(memo["expiry"]))
                    if memo.get("expiry")
                    else None
                ),
                payable_details=memo.get("payableDetails"),
                txn_hash=memo.get("txHash"),
                signed_txn_hash=memo.get("signedTxHash"),
                state=ACPMemoState(memo.get("state")) if memo.get("state") else None,
            )
            for memo in data["memos"]
        ]

        memo_to_sign = (
            next((m for m in memos if int(m.id) == int(memo_to_sign_id)), None)
            if memo_to_sign_id is not None
            else None
        )

        context = data["context"]
        if isinstance(context, str):
            try:
                context = json.loads(context)
            except json.JSONDecodeError:
                context = None

        job = ACPJob(
            acp_client=self,
            id=data["id"],
            client_address=data["clientAddress"],
            provider_address=data["providerAddress"],
            evaluator_address=data["evaluatorAddress"],
            price=data["price"],
            price_token_address=data["priceTokenAddress"],
            memos=memos,
            phase=data["phase"],
            context=context,
            contract_address=data.get("contractAddress"),
            net_payable_amount=data.get("netPayableAmount"),
            deliverable=data.get("deliverable"),
        )
        if self.on_new_task:
            self.on_new_task(job, memo_to_sign)

    def handle_evaluate(self, data) -> None:
        memos = [
            ACPMemo(
                contract_client=self.contract_client_by_address(
                    data.get("contractAddress")
                ),
                id=memo.get("id"),
                type=MemoType(int(memo.get("memoType"))),
                content=memo.get("content"),
                next_phase=ACPJobPhase.from_value(memo.get("nextPhase")),
                status=ACPMemoStatus(memo.get("status")),
                signed_reason=memo.get("signedReason"),
                expiry=(
                    datetime.fromtimestamp(int(memo["expiry"]))
                    if memo.get("expiry")
                    else None
                ),
                payable_details=memo.get("payableDetails"),
                txn_hash=memo.get("txHash"),
                signed_txn_hash=memo.get("signedTxHash"),
                state=ACPMemoState(memo.get("state")) if memo.get("state") else None,
            )
            for memo in data["memos"]
        ]

        context = data["context"]
        if isinstance(context, str):
            try:
                context = json.loads(context)
            except json.JSONDecodeError:
                context = None

        job = ACPJob(
            acp_client=self,
            id=data["id"],
            client_address=data["clientAddress"],
            provider_address=data["providerAddress"],
            evaluator_address=data["evaluatorAddress"],
            price=data["price"],
            price_token_address=data["priceTokenAddress"],
            memos=memos,
            phase=data["phase"],
            context=context,
            contract_address=data.get("contractAddress"),
            net_payable_amount=data.get("netPayableAmount"),
            deliverable=data.get("deliverable"),
        )
        self.on_evaluate(job)

    def __del__(self):
        """Cleanup when the object is destroyed."""
        if hasattr(self, "sio") and self.sio is not None:
            self.sio.disconnect()

    def _hydrate_agent(self, agent_data: Dict[str, Any]) -> IACPAgent:
        contract_address = Web3.to_checksum_address(agent_data.get("contractAddress"))
        if not contract_address:
            raise ACPError("Agent contract address is required")

        contract_client = self.contract_client_by_address(contract_address)
        provider_address = agent_data.get("walletAddress")

        job_offerings: List[ACPJobOffering] = []
        for offering in agent_data.get("jobs", []):
            if "priceV2" in offering:
                price = offering["priceV2"]["value"]
                price_type = PriceType(offering["priceV2"]["type"])
            elif "price" in offering:
                price = offering["price"]
                price_type = PriceType.FIXED
            else:
                continue

            job_offerings.append(
                ACPJobOffering(
                    acp_client=self,
                    contract_client=contract_client,
                    provider_address=provider_address,
                    name=offering["name"],
                    price=price,
                    price_type=price_type,
                    required_funds=offering["requiredFunds"],
                    sla_minutes=offering["slaMinutes"],
                    requirement=offering.get("requirement", None),
                    deliverable=offering.get("deliverable", None),
                )
            )

        resources = [
            ACPResourceOffering(
                acp_client=self,
                name=resource["name"],
                description=resource["description"],
                url=resource["url"],
                parameters=resource.get("parameters", None),
                id=resource["id"],
            )
            for resource in agent_data.get("resources", [])
        ]

        return IACPAgent(
            id=agent_data["id"],
            name=agent_data.get("name"),
            description=agent_data.get("description"),
            wallet_address=Web3.to_checksum_address(agent_data["walletAddress"]),
            job_offerings=job_offerings,
            resources=resources,
            cluster=agent_data.get("cluster"),
            twitter_handle=agent_data.get("twitterHandle"),
            metrics=agent_data.get("metrics"),
            contract_address=contract_address,
        )

    def browse_agents(
        self,
        keyword: str,
        cluster: Optional[str] = None,
        sort_by: Optional[List[ACPAgentSort]] = None,
        top_k: Optional[int] = None,
        graduation_status: Optional[ACPGraduationStatus] = None,
        online_status: Optional[ACPOnlineStatus] = None,
        show_hidden_offerings: bool = False,
    ) -> List[IACPAgent]:
        url = f"{self.acp_url}/agents/v4/search?search={keyword}"
        top_k = 5 if top_k is None else top_k

        if sort_by:
            url += f"&sortBy={','.join([s.value for s in sort_by])}"

        if top_k:
            url += f"&top_k={top_k}"

        if self.wallet_address:
            url += f"&walletAddressesToExclude={self.wallet_address}"

        if cluster:
            url += f"&cluster={cluster}"

        if graduation_status is not None:
            url += f"&graduationStatus={graduation_status.value}"

        if online_status is not None:
            url += f"&onlineStatus={online_status.value}"

        if show_hidden_offerings:
            url += f"&showHiddenOfferings=true"

        try:
            response = requests.get(url)
            response.raise_for_status()
            data = response.json()

            agents_data = data.get("data", [])

            # Filter agents by available contract addresses
            available_contract_addresses = [
                client.contract_address.lower() for client in self.contract_clients
            ]

            # Filter out self and agents not using our contract addresses
            filtered_agents = [
                agent
                for agent in agents_data
                if agent["walletAddress"].lower() != self.wallet_address.lower()
                   and agent.get("contractAddress", "").lower()
                   in available_contract_addresses
            ]

            agents = []
            for agent_data in filtered_agents:
                try:
                    agents.append(self._hydrate_agent(agent_data))
                except Exception as e:
                    logger.warning(f"Failed to hydrate agent {agent_data.get('id')}: {e}")
                    continue

            return agents
        except requests.exceptions.RequestException as e:
            raise ACPApiError(f"Failed to browse agents: {e}")
        except Exception as e:
            raise ACPError(f"An unexpected error occurred while browsing agents: {e}")

    def initiate_job(
        self,
        provider_address: str,
        service_requirement: Union[Dict[str, Any], str],
        fare_amount: FareAmountBase,
        evaluator_address: Optional[str] = None,
        expired_at: Optional[datetime] = None,
    ) -> int:
        if expired_at is None:
            expired_at = datetime.now(timezone.utc) + timedelta(days=1)

        if provider_address == self.wallet_address:
            raise ACPError("Provider address cannot be the same as the client address")

        eval_addr = (
            Web3.to_checksum_address(evaluator_address)
            if evaluator_address
            else self.wallet_address
        )

        # Lookup existing account between client and provider
        account = self.get_by_client_and_provider(
            self.wallet_address, provider_address, self.acp_contract_client
        )

        # Determine whether to call createJob or createJobWithAccount
        base_contract_addresses = {
            BASE_SEPOLIA_CONFIG.contract_address.lower(),
            BASE_SEPOLIA_ACP_X402_CONFIG.contract_address.lower(),
            BASE_MAINNET_CONFIG.contract_address.lower(),
            BASE_MAINNET_ACP_X402_CONFIG.contract_address.lower(),

        }

        use_simple_create = (
            self.acp_contract_client.config.contract_address.lower()
            in base_contract_addresses
        )

        chain_id = self.acp_contract_client.config.chain_id
        usdc_token_address = USDC_TOKEN_ADDRESS[chain_id]
        is_usdc_payment_token = usdc_token_address == fare_amount.fare.contract_address
        is_x402_job = bool(getattr(self.acp_contract_client.config, "x402_config", None) and is_usdc_payment_token)

        if use_simple_create or not account:
            create_job_operation = self.acp_contract_client.create_job(
                provider_address,
                eval_addr or self.wallet_address,
                expired_at,
                fare_amount.fare.contract_address,
                fare_amount.amount,
                "",
                is_x402_job=is_x402_job,
            )
        else:
            create_job_operation = self.acp_contract_client.create_job_with_account(
                account.id,
                eval_addr or self.wallet_address,
                fare_amount.amount,
                fare_amount.fare.contract_address,
                expired_at,
                is_x402_job=is_x402_job,
            )

        response = self.acp_contract_client.handle_operation([create_job_operation])

        job_id = self.acp_contract_client.get_job_id(
            response, self.wallet_address, provider_address
        )

        operations = self.acp_contract_client.create_memo(
            job_id,
            (
                service_requirement
                if isinstance(service_requirement, str)
                else json.dumps(service_requirement)
            ),
            MemoType.MESSAGE,
            is_secured=True,
            next_phase=ACPJobPhase.NEGOTIATION,
        )

        self.acp_contract_client.handle_operation([operations])

        return job_id

    def get_by_client_and_provider(
        self,
        client_address: str,
        provider_address: str,
        acp_contract_client: Optional[BaseAcpContractClient] = None,
    ) -> Optional[ACPAccount]:
        """Get account by client and provider addresses."""
        try:
            url = f"{self.acp_url}/accounts/client/{client_address}/provider/{provider_address}"

            response = requests.get(url)
            if response.status_code == 404:
                return None

            response.raise_for_status()
            data = response.json()

            if not data.get("data"):
                return None

            account_data = data["data"]
            contract_client = acp_contract_client or self.contract_clients[0]

            return ACPAccount(
                contract_client=contract_client,
                id=account_data["id"],
                client_address=account_data["clientAddress"],
                provider_address=account_data["providerAddress"],
                metadata=account_data.get("metadata", ""),
            )
        except requests.exceptions.RequestException as e:
            raise ACPApiError(f"Failed to get account by client and provider: {e}")
        except Exception as e:
            raise ACPError(f"An unexpected error occurred while getting account: {e}")

    def get_account_by_job_id(
        self,
        job_id: int,
        acp_contract_client: Optional[BaseAcpContractClient] = None,
    ) -> Optional[ACPAccount]:
        """Get account by job ID."""
        try:
            url = f"{self.acp_url}/accounts/job/{job_id}"

            response = requests.get(url)
            response.raise_for_status()
            data = response.json()

            if not data.get("data"):
                return None

            account_data = data["data"]
            contract_client = acp_contract_client or self.contract_clients[0]

            return ACPAccount(
                contract_client=contract_client,
                id=account_data["id"],
                client_address=account_data["clientAddress"],
                provider_address=account_data["providerAddress"],
                metadata=account_data.get("metadata", ""),
            )
        except requests.exceptions.RequestException as e:
            raise ACPApiError(f"Failed to get account by job id: {e}")
        except Exception as e:
            raise ACPError(
                f"An unexpected error occurred while getting account by job id: {e}"
            )

    def get_active_jobs(self, page: int = 1, page_size: int = 10) -> List["ACPJob"]:
        url = f"{self.acp_url}/jobs/active?pagination[page]={page}&pagination[pageSize]={page_size}"
        raw_jobs = self._fetch_job_list(url)
        return self._hydrate_jobs(raw_jobs, log_prefix="Active jobs")

    def get_pending_memo_jobs(self, page: int = 1, page_size: int = 10) -> List["ACPJob"]:
        url = f"{self.acp_url}/jobs/pending-memos?pagination[page]={page}&pagination[pageSize]={page_size}"
        raw_jobs = self._fetch_job_list(url)
        return self._hydrate_jobs(raw_jobs, log_prefix="Pending memo jobs")

    def get_completed_jobs(self, page: int = 1, page_size: int = 10) -> List["ACPJob"]:
        url = f"{self.acp_url}/jobs/completed?pagination[page]={page}&pagination[pageSize]={page_size}"
        raw_jobs = self._fetch_job_list(url)
        return self._hydrate_jobs(raw_jobs, log_prefix="Completed jobs")

    def get_cancelled_jobs(self, page: int = 1, page_size: int = 10) -> List["ACPJob"]:
        url = f"{self.acp_url}/jobs/cancelled?pagination[page]={page}&pagination[pageSize]={page_size}"
        raw_jobs = self._fetch_job_list(url)
        return self._hydrate_jobs(raw_jobs, log_prefix="Cancelled jobs")

    def _fetch_job_list(
        self,
        url: str,
    ) -> List[dict]:
        try:
            response = requests.get(
                url,
                headers={"wallet-address": self.wallet_address},
            )
            response.raise_for_status()
        except requests.RequestException as e:
            raise ACPApiError("Failed to fetch ACP jobs (network error)") from e

        try:
            data = response.json()
        except ValueError as e:
            raise ACPApiError("Failed to parse ACP jobs response") from e

        if data.get("error"):
            raise ACPApiError(data["error"]["message"])

        return data.get("data", [])

    def _hydrate_jobs(
        self,
        raw_jobs: List[dict],
        *,
        log_prefix: str = "Skipped",
    ) -> List[ACPJob]:
        jobs: List[ACPJob] = []
        errors: list[dict] = []

        for job in raw_jobs:
            try:
                memos = [
                    ACPMemo(
                        contract_client=self.contract_client_by_address(
                            job.get("contractAddress")
                        ),
                        id=memo.get("id"),
                        type=MemoType(int(memo.get("memoType"))),
                        content=memo.get("content"),
                        next_phase=ACPJobPhase.from_value(memo.get("nextPhase")),
                        status=ACPMemoStatus(memo.get("status")),
                        signed_reason=memo.get("signedReason"),
                        expiry=(
                            datetime.fromtimestamp(int(memo["expiry"]))
                            if memo.get("expiry")
                            else None
                        ),
                        payable_details=memo.get("payableDetails"),
                        txn_hash=memo.get("txHash"),
                        signed_txn_hash=memo.get("signedTxHash"),
                        state=ACPMemoState(memo.get("state")) if memo.get("state") else None,
                    )
                    for memo in job.get("memos", [])
                ]

                context = job.get("context")
                if isinstance(context, str):
                    try:
                        context = json.loads(context)
                    except json.JSONDecodeError:
                        context = None

                jobs.append(
                    ACPJob(
                        acp_client=self,
                        id=job.get("id"),
                        client_address=job.get("clientAddress"),
                        provider_address=job.get("providerAddress"),
                        evaluator_address=job.get("evaluatorAddress"),
                        price=job.get("price"),
                        price_token_address=job.get("priceTokenAddress"),
                        memos=memos,
                        phase=job.get("phase"),
                        context=context,
                        contract_address=job.get("contractAddress"),
                        net_payable_amount=job.get("netPayableAmount"),
                        deliverable=job.get("deliverable"),
                    )
                )

            except Exception as e:
                errors.append(
                    {
                        "job_id": job.get("id"),
                        "error": e,
                    }
                )

            if errors:
                payload = [
                    {
                        "job_id": e["job_id"],
                        "message": str(e["error"]),
                    }
                    for e in errors
                ]

                logger.warning(
                    "[ACP] %s %d malformed job(s):\n%s",
                    log_prefix,
                    len(errors),
                    json.dumps(payload, indent=2),
                )

        return jobs

    def get_job_by_onchain_id(self, onchain_job_id: int) -> "ACPJob":
        url = f"{self.acp_url}/jobs/{onchain_job_id}"
        headers = {"wallet-address": self.wallet_address}

        try:
            response = requests.get(url, headers=headers)
            response.raise_for_status()
            data = response.json()

            if data.get("error"):
                raise ACPApiError(data["error"]["message"])

            memos = []
            for memo in data.get("data", {}).get("memos", []):
                memos.append(
                    ACPMemo(
                        contract_client=self.acp_contract_client,
                        id=memo.get("id"),
                        type=MemoType(int(memo.get("memoType"))),
                        content=memo.get("content"),
                        next_phase=ACPJobPhase.from_value(memo.get("nextPhase")),
                        status=ACPMemoStatus(memo.get("status")),
                        signed_reason=memo.get("signedReason"),
                        expiry=(
                            datetime.fromtimestamp(int(memo["expiry"]))
                            if memo.get("expiry")
                            else None
                        ),
                        payable_details=memo.get("payableDetails"),
                        txn_hash=memo.get("txHash"),
                        signed_txn_hash=memo.get("signedTxHash"),
                        state=ACPMemoState(memo.get("state")) if memo.get("state") else None,
                    )
                )

            context = data.get("data", {}).get("context")
            if isinstance(context, str):
                try:
                    context = json.loads(context)
                except json.JSONDecodeError:
                    context = None

            job = data.get("data", {})
            return ACPJob(
                acp_client=self,
                id=job["id"],
                client_address=job["clientAddress"],
                provider_address=job["providerAddress"],
                evaluator_address=job["evaluatorAddress"],
                price=job["price"],
                price_token_address=job["priceTokenAddress"],
                memos=memos,
                phase=job["phase"],
                context=context,
                contract_address=job.get("contractAddress"),
                net_payable_amount=job.get("netPayableAmount"),
                deliverable=job.get("deliverable"),
            )
        except Exception as e:
            raise ACPApiError(f"Failed to get job by onchain ID: {e}")

    def get_memo_by_id(self, onchain_job_id: int, memo_id: int) -> "ACPMemo":
        url = f"{self.acp_url}/jobs/{onchain_job_id}/memos/{memo_id}"
        headers = {"wallet-address": self.wallet_address}

        try:
            response = requests.get(url, headers=headers)
            response.raise_for_status()
            data = response.json()

            if data.get("error"):
                raise ACPApiError(data["error"]["message"])

            memo = data.get("data", {})

            return ACPMemo(
                contract_client=self.acp_contract_client,
                id=memo.get("id"),
                type=MemoType(memo.get("memoType")),
                content=memo.get("content"),
                next_phase=ACPJobPhase.from_value(memo.get("nextPhase")),
                status=ACPMemoStatus(memo.get("status")),
                signed_reason=memo.get("signedReason"),
                expiry=(
                    datetime.fromtimestamp(int(memo["expiry"]))
                    if memo.get("expiry")
                    else None
                ),
                payable_details=memo.get("payableDetails"),
                txn_hash=memo.get("txHash"),
                signed_txn_hash=memo.get("signedTxHash"),
                state=ACPMemoState(memo.get("state")) if memo.get("state") else None,
            )

        except Exception as e:
            raise ACPApiError(f"Failed to get memo by ID: {e}")

    def get_agent(self, wallet_address: str, *, show_hidden_offerings: bool = False) -> Optional[IACPAgent]:
        url = f"{self.acp_url}/agents?filters[walletAddress]={wallet_address}"

        if show_hidden_offerings:
            url += f"&showHiddenOfferings=true"

        try:
            response = requests.get(url)
            response.raise_for_status()
            data = response.json()

            agents_data = data.get("data", [])
            if not agents_data:
                return None

            agent_data = agents_data[0]
            return self._hydrate_agent(agent_data)

        except requests.exceptions.RequestException as e:
            raise ACPApiError(f"Failed to get agent: {e}")
        except Exception as e:
            raise ACPError(f"An unexpected error occurred while getting agent: {e}")

    def get_memo_content(self, url: str) -> str:
        response = self.acp_client.request("GET", url)

        if not response:
            raise ACPApiError("Failed to get memo content")

        return response["content"]


# Rebuild the AcpJob model after VirtualsACP is defined
ACPJob.model_rebuild()
ACPMemo.model_rebuild()
ACPJobOffering.model_rebuild()
ACPResourceOffering.model_rebuild()



import json
import logging
import math
import secrets
from datetime import datetime
from typing import Dict, Any, Optional, List

from eth_account import Account
from eth_account.messages import encode_typed_data
from eth_utils.crypto import keccak
from web3 import Web3

from virtuals_acp.alchemy import AlchemyAccountKit
from virtuals_acp.configs.configs import ACPContractConfig, BASE_MAINNET_CONFIG
from virtuals_acp.contract_clients.base_contract_client import BaseAcpContractClient
from virtuals_acp.constants import SINGLE_SIGNER_VALIDATION_MODULE_ADDRESS
from virtuals_acp.exceptions import ACPError
from virtuals_acp.models import (
    ACPJobPhase,
    MemoType,
    FeeType,
    X402PayableRequest,
    X402Payment,
    X402PayableRequirements,
    OperationPayload,
    OffChainJob,
)
from virtuals_acp.x402 import ACPX402

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
)
logger = logging.getLogger("ContractClient")

class ACPContractClient(BaseAcpContractClient):
    def __init__(
        self,
        wallet_private_key: str,
        agent_wallet_address: str,
        entity_id: int,
        config: ACPContractConfig = BASE_MAINNET_CONFIG,
    ):
        super().__init__(agent_wallet_address, config)
        self.account = Account.from_key(wallet_private_key)
        self.entity_id = entity_id
        self.alchemy_kit = AlchemyAccountKit(
            config, agent_wallet_address, entity_id, self.account, config.chain_id
        )
        self.x402 = ACPX402(
            config, self.account, self.w3, self.agent_wallet_address, self.entity_id
        )

        self.validate_session_key_on_chain(self.account.address, self.entity_id)

        logger.info(
            "\nConnected to ACP:\n%s",
            json.dumps(
                {
                    "agent_wallet_address": agent_wallet_address,
                    "whitelisted_wallet_address": self.account.address,
                    "entity_id": self.entity_id,
                },
                indent=2
            )
        )
        
    def get_acp_version(self) -> str:
        return "1"

    def _get_random_nonce(self, bits: int = 152) -> int:
        """Generate a random bigint nonce."""
        bytes_len = bits // 8
        random_bytes = secrets.token_bytes(bytes_len)
        return int.from_bytes(random_bytes, byteorder="big")

    def handle_operation(self, trx_data: List[OperationPayload]) -> Dict[str, Any]:
        return self.alchemy_kit.handle_user_operation(trx_data)

    def get_job_id(
        self, response: Dict[str, Any], client_address: str, provider_address: str
    ) -> int:
        logs: List[Dict[str, Any]] = response.get("receipts", [])[0].get("logs", [])

        decoded_create_job_logs = [
            self.contract.events.JobCreated().process_log(
                {
                    "topics": log["topics"],
                    "data": log["data"],
                    "address": log["address"],
                    "logIndex": 0,
                    "transactionIndex": 0,
                    "transactionHash": "0x0000",
                    "blockHash": "0x0000",
                    "blockNumber": 0,
                }
            )
            for log in logs
            if log["topics"][0] == self.job_created_event_signature_hex
        ]

        if len(decoded_create_job_logs) == 0:
            raise Exception("No logs found for JobCreated event")

        created_job_log = next(
            (
                log
                for log in decoded_create_job_logs
                if log["args"]["provider"] == provider_address
                and log["args"]["client"] == client_address
            ),
            None,
        )

        if not created_job_log:
            raise Exception(
                "No logs found for JobCreated event with provider and client addresses"
            )

        return int(created_job_log["args"]["jobId"])

    def create_job(
        self,
        provider_address: str,
        evaluator_address: str,
        expire_at: datetime,
        payment_token_address: str,
        budget_base_unit: int,
        metadata: str = "",
        is_x402_job: bool = False
    ) -> OperationPayload:
        try:
            provider_address = Web3.to_checksum_address(provider_address)
            evaluator_address = Web3.to_checksum_address(evaluator_address)
            expire_timestamp = math.floor(expire_at.timestamp())
            
            fn_name = "createJobWithX402" if is_x402_job else "createJob"

            operation = self._build_user_operation(
                fn_name, [provider_address, evaluator_address, expire_timestamp]
            )

            return OperationPayload(
                data=operation["data"],
                to=operation["to"],
            )
        except Exception as e:
            raise ACPError("Failed to create job", e)

    def set_budget_with_payment_token(
        self,
        job_id: int,
        budget_base_unit: int,
        payment_token_address: Optional[str] = None,
    ) -> OperationPayload:
        token = payment_token_address or self.config.base_fare.contract_address
        operation = self._build_user_operation(
            "setBudgetWithPaymentToken", [job_id, budget_base_unit, token]
        )

        return OperationPayload(
            data=operation["data"],
            to=operation["to"],
        )

    def create_payable_memo(
        self,
        job_id: int,
        content: str,
        amount_base_unit: int,
        recipient: str,
        fee_amount_base_unit: int,
        fee_type: FeeType,
        next_phase: ACPJobPhase,
        memo_type: MemoType,
        expired_at: datetime,
        token: Optional[str] = None,
        secured: bool = True,
    ) -> OperationPayload:
        try:
            token_address = token or self.config.base_fare.contract_address
            operation = self._build_user_operation(
                "createPayableMemo",
                [
                    job_id,
                    content,
                    token_address,
                    amount_base_unit,
                    Web3.to_checksum_address(recipient),
                    fee_amount_base_unit,
                    fee_type.value,
                    memo_type.value,
                    next_phase.value,
                    math.floor(expired_at.timestamp()),
                ],
            )

            return OperationPayload(
                data=operation["data"],
                to=operation["to"],
            )
        except Exception as e:
            raise ACPError("Failed to create payable memo", e)

    def create_job_with_account(
        self,
        account_id: int,
        evaluator_address: str,
        budget_base_unit: int,
        payment_token_address: str,
        expired_at: datetime,
    ) -> Dict[str, Any]:
        raise ACPError("Not Supported")

    def update_account_metadata(self, account_id: int, metadata: str) -> Dict[str, Any]:
        raise ACPError("Not Supported")

    def update_job_x402_nonce(self, job_id: int, nonce: str) -> OffChainJob:
        """Update job X402 nonce."""
        try:
            return self.x402.update_job_nonce(job_id, nonce)
        except Exception as e:
            raise ACPError("Failed to update job X402 nonce", e)

    def generate_x402_payment(
        self, payable_request: X402PayableRequest, requirements: X402PayableRequirements
    ) -> X402Payment:
        """Generate X402 payment."""
        try:
            return self.x402.generate_payment(payable_request, requirements)
        except Exception as e:
            raise ACPError("Failed to generate X402 payment", e)

    def perform_x402_request(
        self, url: str, version: str, budget: Optional[str] = None, signature: Optional[str] = None
    ) -> Dict[str, Any]:
        try:
            return self.x402.perform_request(url, version, budget, signature)
        except Exception as e:
            raise ACPError("Failed to perform X402 request", e)

    def get_asset_manager_address(self) -> str:
        raise ACPError("Not Supported")

    def sign_typed_data(self, typed_data: dict[str, Any]) -> str:
        encoded = encode_typed_data(full_message=typed_data)
        typed_data_hash = keccak(b"\x19\x01" + encoded.header + encoded.body)

        replay_safe_typed_data = {
            "domain": {
                "chainId": self.config.chain_id,
                "verifyingContract": SINGLE_SIGNER_VALIDATION_MODULE_ADDRESS,
                "salt": "0x" + "00" * 12 + self.agent_wallet_address[2:],
            },
            "types": {"ReplaySafeHash": [{"name": "hash", "type": "bytes32"}]},
            "message": {"hash": "0x" + typed_data_hash.hex()},
            "primaryType": "ReplaySafeHash",
        }

        signable = encode_typed_data(full_message=replay_safe_typed_data)
        signed = self.account.sign_message(signable)
        raw_signature = signed.signature.hex()
        return self._pack_1271_eoa_signature(raw_signature)

    def _pack_1271_eoa_signature(self, validation_signature: str) -> str:
        if validation_signature.startswith("0x"):
            validation_signature = validation_signature[2:]

        prefix = b"\x00"
        entity_id_bytes = self.entity_id.to_bytes(4, "big")
        separator = b"\xff"
        eoa_type = b"\x00"
        sig_bytes = bytes.fromhex(validation_signature)

        packed = prefix + entity_id_bytes + separator + eoa_type + sig_bytes
        return packed.hex()
